import uuid

from edkrule.engine.constant import Constant
from edkrule.engine.engine import Engine
from edkrule.engine.finder.finder import Finder
from edkrule.interpreter.expression.categorize import Categorize
from edkrule.interpreter.lexer.token_type import TokenType
from edkrule.interpreter.lexer.tokens.token import Token


class Expression:
    def __init__(self, body=None):
        if body is None: body = []
        self._body = body
        self._name = None
        self._engine = None
        self.rid = uuid.uuid4().hex
        self._origin_statement = None

    def origin_statement(self, statement):
        self._origin_statement = statement

    def append(self, body):
        self._body.append(body)

    def insert(self, body):
        self._body.insert(0, body)

    @property
    def engine(self) -> Engine:
        return self._engine

    @engine.setter
    def engine(self, value):
        self._engine = value
        for body in self._body:
            if type(body) != Token:
                body.engine = value

    def run(self):
        if self._engine is None: pass
        if self.name == "Anonymous":
            return self.anonymous()
        if self.name == 'TernaryOperator':
            return self.ternary()
        return self.func()

    def func(self):
        runner = self.engine.get_class(Constant.Expressions, self.name)(self.engine, self.rid)
        return runner.run(*[b.run() for b in self._body])

    def anonymous(self):
        if len(self._body) == 1:
            return self._body[0].run()
        else:
            return self.binary()

    def binary(self):
        runner = self.engine.get_class(Constant.Expressions, self._body[1].text)(self.engine, self.rid)
        return runner.run(*[self.body[0].run(), self.body[2].run()])

    def ternary(self):
        if self._body[0].run():
            return self._body[2].run()
        else:
            return self._body[4].run()

    @property
    def type(self):
        return TokenType.Expression

    @property
    def name(self):
        if self._name is None: return "Anonymous"
        return self._name.text

    @name.setter
    def name(self, value):
        self._name = value

    def display2(self, index):
        """
        测试用
        :param index:
        :type index:
        :return:
        :rtype:
        """
        if not index: print(self.name)
        for i, e in enumerate(self._body):
            pos = []
            pos.extend(index)
            pos.append(str(i))
            if Categorize.is_variable(e) or Categorize.is_op(e):
                # exp.matrix(0).name == 'Anonymous'
                print("assert", f'exp.matrix({",".join(pos)}).text == \'{e.text}\'')
                print("assert", f'exp.matrix({",".join(pos)}).type == {e.type}')
            else:
                print("assert", f'exp.matrix({",".join(pos)}).name == \'{e.name}\'')
                print("assert", f'exp.matrix({",".join(pos)}).type == {e.type}')
                # print(",".join(pos), e.name, e.type)
                e.display2(pos)

    def tree_data(self):
        tree_json = dict()
        tree_json["name"] = self.name
        tree_json["children"] = []
        tree_json["value"] = self.text
        tree_json["type"] = "statement"
        for i, e in enumerate(self._body):
            if Categorize.is_variable(e) or Categorize.is_op(e):
                tree_json["children"].append(dict(name=e.text, value=e.text, type=e.type.value))
            else:
                tree_json["children"].append(e.tree_data())
        return tree_json

    def go_tree_data(self, parent_id=None):
        result = []
        tree_json = dict()
        tree_json["name"] = self.text
        tree_json["id"] = str(uuid.uuid4())
        tree_json["gender"] = "F"
        if parent_id: tree_json["parent"] = parent_id
        result.append(tree_json)
        for i, e in enumerate(self._body):
            if e.type == TokenType.Variable:
                result.append(dict(name=e.text, gender="M",parent=tree_json["id"], id=str(uuid.uuid4())))
            elif e.type == TokenType.Expression:
                result.extend(e.go_tree_data(tree_json["id"]))
        return result

    def display(self, level: int = 0, indexes=None):
        """
        测试用
        :param index:
        :type index:
        :return:
        :rtype:
        """
        if indexes is None: indexes = list()
        print(self.name)
        for index, e in enumerate(self._body):
            if Categorize.is_variable(e) or Categorize.is_op(e):
                print(e, e.type, e.text)
            else:
                print(f'{self.blank(level)}------{e.name}-----Start')
                e.display(level + 1, indexes)
                print(f'{self.blank(level)}------{e.name}-----End')

    def blank(self, level):
        return "".join([" " for i in range(level + 1)])

    @property
    def body(self):
        return self._body

    def matrix(self, *args):
        deep = len(args)
        index = 0
        result = self.body[args[index]]
        while deep > 1:
            deep -= 1
            index += 1
            result = result.body[args[index]]
        return result

    def find(self, finder: Finder, finder_parameter):
        return finder.find(self, finder_parameter)

    @property
    def text(self):
        if self._origin_statement is None:
            if len(self.body) == 1 and not self.is_function():
                return self.body[0].text
            if not self.is_function():
                return "".join([e.text for e in self.body])
            else:
                return f"{self.name}({','.join([e.text for e in self.body])})"
        else:
            return self._origin_statement.text

    def is_function(self):
        return self.name != "Anonymous" and self.name != "TernaryOperator"



from edkrule.interpreter.lexer.character import Character
from edkrule.interpreter.lexer.dfa_state import DfaState
from edkrule.interpreter.lexer.token_type import TokenType


class LtToken:
    @staticmethod
    def accept(lexer, char: str):
        if Character.isless(char):
            lexer.state = DfaState.Lt
            lexer.token_type = TokenType.Lt
            lexer.token_text += char
            return True
        return False



from edkrule.interpreter.lexer.character import Character
from edkrule.interpreter.lexer.dfa_state import DfaState
from edkrule.interpreter.lexer.token_type import TokenType


class VariableState:
    # dot_numer = 0
    # next_dot = False

    @staticmethod
    def accept(lexer, char: str, state: DfaState):
        if state == DfaState.Var:
            if lexer.dot_numer > 1:
                if Character.isrightparentheses(char) or Character.iscomma(char) or (
                        Character.isoperate(char) and lexer.next_dot is False):
                    lexer.token_type = TokenType.Variable
                    lexer.dot_numer = 0
                    lexer.init_token(char)
                else: lexer.token_text += char

                lexer.next_dot = False
            else:
                if Character.isdot(char):
                    lexer.dot_numer += 1
                    lexer.next_dot = True
                else: lexer.next_dot = False
                lexer.token_text += char
            return True
        return False



from typing import Any


class Stack:
    def __init__(self):
        self._stack = []

    def insert(self, e):
        self._stack.insert(0, e)

    def push(self, e):
        self._stack.append(e)

    def pop(self):
        return self._stack.pop()

    def clear(self):
        self._stack.clear()

    def pops(self, count):
        pops = []
        while count > 0:
            pops.insert(0, self.pop())
            count -= 1
        return pops

    def dequeues(self, count):
        dequeues = []
        while count > 0:
            dequeues.append(self._stack.pop(0))
            count -= 1
        return dequeues

    def dequeue(self):
        return self._stack.pop(0)


    def pop_until(self, condition: Any, exclude_condition=True):
        pops = []
        while True:
            # pops.insert(0, self.pop())
            # if self.empty(): break
            if condition(self.top()):
                pops.insert(0, self.pop())
                if exclude_condition:
                    pops.pop(0)
                break
            else:
                pops.insert(0, self.pop())

        return pops

    def top(self, index: int = 0):
        if len(self._stack) == 0: return None
        return self._stack[index-1]

    def count(self):
        return len(self._stack)

    def empty(self):
        return self.count() == 0


import json

from edkrule.edk_rule import EdkRule


def out_put(data):
    with open("debug.json", 'w') as f:
        json.dump(data, f)

def test_case1():
    rule_string = """a==1&&b==1?true:false"""
    expression = EdkRule.expression(rule_string)
    data = expression.tree_data()
    EdkRule.draw(rule_string, "rule1.html")
    # out_put(data)

    assert data == {"name": "TernaryOperator", "children": [{"name": "Anonymous", "children": [{"name": "Anonymous", "children": [{"name": "a", "value": "a", "type": 1}, {"name": "==", "value": "==", "type": 3}, {"name": "1", "value": "1", "type": 17}], "value": "a==1", "type": "statement"}, {"name": "&&", "value": "&&", "type": 19}, {"name": "Anonymous", "children": [{"name": "b", "value": "b", "type": 1}, {"name": "==", "value": "==", "type": 3}, {"name": "1", "value": "1", "type": 17}], "value": "b==1", "type": "statement"}], "value": "a==1&&b==1", "type": "statement"}, {"name": "?", "value": "?", "type": 30}, {"name": "true", "value": "true", "type": 22}, {"name": ":", "value": ":", "type": 14}, {"name": "false", "value": "false", "type": 23}], "value": "a==1&&b==1?true:false", "type": "statement"}



def test_case2():
    rule_string = """a&&b==1?x==1?true:false:false"""
    expression = EdkRule.expression(rule_string)
    data = expression.tree_data()
    EdkRule.draw(rule_string, "rule2.html")
    out_put(data)
    assert  data == {"name": "TernaryOperator", "children": [{"name": "Anonymous", "children": [{"name": "a", "value": "a", "type": 1}, {"name": "&&", "value": "&&", "type": 19}, {"name": "Anonymous", "children": [{"name": "b", "value": "b", "type": 1}, {"name": "==", "value": "==", "type": 3}, {"name": "1", "value": "1", "type": 17}], "value": "b==1", "type": "statement"}], "value": "a&&b==1", "type": "statement"}, {"name": "?", "value": "?", "type": 30}, {"name": "TernaryOperator", "children": [{"name": "Anonymous", "children": [{"name": "Anonymous", "children": [{"name": "x", "value": "x", "type": 1}, {"name": "==", "value": "==", "type": 3}, {"name": "1", "value": "1", "type": 17}], "value": "x==1", "type": "statement"}], "value": "x==1", "type": "statement"}, {"name": "?", "value": "?", "type": 30}, {"name": "true", "value": "true", "type": 22}, {"name": ":", "value": ":", "type": 14}, {"name": "false", "value": "false", "type": 23}], "value": "x==1?true:false", "type": "statement"}, {"name": ":", "value": ":", "type": 14}, {"name": "false", "value": "false", "type": 23}], "value": "a&&b==1?x==1?true:false:false", "type": "statement"}


def test_case3():
    rule_string = """a&&b==1?x==1?true:false:y==1?true:false"""
    expression = EdkRule.expression(rule_string)
    data = expression.tree_data()
    EdkRule.draw(rule_string, "rule3.html")
    out_put(data)
    assert data == {"name": "TernaryOperator", "children": [{"name": "Anonymous", "children": [{"name": "a", "value": "a", "type": 1}, {"name": "&&", "value": "&&", "type": 19}, {"name": "Anonymous", "children": [{"name": "b", "value": "b", "type": 1}, {"name": "==", "value": "==", "type": 3}, {"name": "1", "value": "1", "type": 17}], "value": "b==1", "type": "statement"}], "value": "a&&b==1", "type": "statement"}, {"name": "?", "value": "?", "type": 30}, {"name": "TernaryOperator", "children": [{"name": "Anonymous", "children": [{"name": "Anonymous", "children": [{"name": "x", "value": "x", "type": 1}, {"name": "==", "value": "==", "type": 3}, {"name": "1", "value": "1", "type": 17}], "value": "x==1", "type": "statement"}], "value": "x==1", "type": "statement"}, {"name": "?", "value": "?", "type": 30}, {"name": "true", "value": "true", "type": 22}, {"name": ":", "value": ":", "type": 14}, {"name": "false", "value": "false", "type": 23}], "value": "x==1?true:false", "type": "statement"}, {"name": ":", "value": ":", "type": 14}, {"name": "TernaryOperator", "children": [{"name": "Anonymous", "children": [{"name": "Anonymous", "children": [{"name": "y", "value": "y", "type": 1}, {"name": "==", "value": "==", "type": 3}, {"name": "1", "value": "1", "type": 17}], "value": "y==1", "type": "statement"}], "value": "y==1", "type": "statement"}, {"name": "?", "value": "?", "type": 30}, {"name": "true", "value": "true", "type": 22}, {"name": ":", "value": ":", "type": 14}, {"name": "false", "value": "false", "type": 23}], "value": "y==1?true:false", "type": "statement"}], "value": "a&&b==1?x==1?true:false:y==1?true:false", "type": "statement"}



def test_case4():
    rule_string = """a==1||b==2||c==3"""
    expression = EdkRule.expression(rule_string)
    data = expression.tree_data()
    EdkRule.draw(rule_string, "rule4.html")
    assert data == {"name": "Anonymous", "children": [{"name": "Anonymous", "children": [{"name": "Anonymous", "children": [{"name": "a", "value": "a", "type": 1}, {"name": "==", "value": "==", "type": 3}, {"name": "1", "value": "1", "type": 17}], "value": "a==1", "type": "statement"}, {"name": "||", "value": "||", "type": 21}, {"name": "Anonymous", "children": [{"name": "b", "value": "b", "type": 1}, {"name": "==", "value": "==", "type": 3}, {"name": "2", "value": "2", "type": 17}], "value": "b==2", "type": "statement"}], "value": "a==1||b==2", "type": "statement"}, {"name": "||", "value": "||", "type": 21}, {"name": "Anonymous", "children": [{"name": "c", "value": "c", "type": 1}, {"name": "==", "value": "==", "type": 3}, {"name": "3", "value": "3", "type": 17}], "value": "c==3", "type": "statement"}], "value": "a==1||b==2||c==3", "type": "statement"}



def test_case5():
    rule_string = """a==1||b==2||c"""
    expression = EdkRule.expression(rule_string)
    data = expression.tree_data()
    EdkRule.draw(rule_string, "rule5.html")
    out_put(data)
    assert data == {"name": "Anonymous", "children": [{"name": "Anonymous", "children": [{"name": "Anonymous", "children": [{"name": "a", "value": "a", "type": 1}, {"name": "==", "value": "==", "type": 3}, {"name": "1", "value": "1", "type": 17}], "value": "a==1", "type": "statement"}, {"name": "||", "value": "||", "type": 21}, {"name": "Anonymous", "children": [{"name": "b", "value": "b", "type": 1}, {"name": "==", "value": "==", "type": 3}, {"name": "2", "value": "2", "type": 17}], "value": "b==2", "type": "statement"}], "value": "a==1||b==2", "type": "statement"}, {"name": "||", "value": "||", "type": 21}, {"name": "c", "value": "c", "type": 1}], "value": "a==1||b==2||c", "type": "statement"}



def test_case6():
    rule_string = """a||b==2||c==2"""
    expression = EdkRule.expression(rule_string)
    data = expression.tree_data()
    EdkRule.draw(rule_string, "rule6.html")
    out_put(data)
    assert data == {"name": "Anonymous", "children": [{"name": "Anonymous", "children": [{"name": "a", "value": "a", "type": 1}, {"name": "||", "value": "||", "type": 21}, {"name": "Anonymous", "children": [{"name": "b", "value": "b", "type": 1}, {"name": "==", "value": "==", "type": 3}, {"name": "2", "value": "2", "type": 17}], "value": "b==2", "type": "statement"}], "value": "a||b==2", "type": "statement"}, {"name": "||", "value": "||", "type": 21}, {"name": "Anonymous", "children": [{"name": "c", "value": "c", "type": 1}, {"name": "==", "value": "==", "type": 3}, {"name": "2", "value": "2", "type": 17}], "value": "c==2", "type": "statement"}], "value": "a||b==2||c==2", "type": "statement"}



def test_case7():
    rule_string = """a||(b==2||c==2)"""
    expression = EdkRule.expression(rule_string)
    data = expression.tree_data()
    EdkRule.draw(rule_string, "rule7.html")
    out_put(data)
    assert data == {"name": "Anonymous", "children": [{"name": "a", "value": "a", "type": 1}, {"name": "||", "value": "||", "type": 21}, {"name": "Anonymous", "children": [{"name": "Anonymous", "children": [{"name": "Anonymous", "children": [{"name": "b", "value": "b", "type": 1}, {"name": "==", "value": "==", "type": 3}, {"name": "2", "value": "2", "type": 17}], "value": "b==2", "type": "statement"}, {"name": "||", "value": "||", "type": 21}, {"name": "Anonymous", "children": [{"name": "c", "value": "c", "type": 1}, {"name": "==", "value": "==", "type": 3}, {"name": "2", "value": "2", "type": 17}], "value": "c==2", "type": "statement"}], "value": "b==2||c==2", "type": "statement"}], "value": "b==2||c==2", "type": "statement"}], "value": "a||(b==2||c==2)", "type": "statement"}



def test_case8():
    rule_string = """max=a>b?a:b"""
    expression = EdkRule.expression(rule_string)
    data = expression.tree_data()
    EdkRule.draw(rule_string, "rule8.html")
    out_put(data)
    assert data == {"name": "Anonymous", "children": [{"name": "max", "value": "max", "type": 1}, {"name": "=", "value": "=", "type": 2}, {"name": "TernaryOperator", "children": [{"name": "Anonymous", "children": [{"name": "Anonymous", "children": [{"name": "a", "value": "a", "type": 1}, {"name": ">", "value": ">", "type": 7}, {"name": "b", "value": "b", "type": 1}], "value": "a>b", "type": "statement"}], "value": "a>b", "type": "statement"}, {"name": "?", "value": "?", "type": 30}, {"name": "a", "value": "a", "type": 1}, {"name": ":", "value": ":", "type": 14}, {"name": "b", "value": "b", "type": 1}], "value": "a>b?a:b", "type": "statement"}], "value": "max=a>b?a:b", "type": "statement"}



def test_case9():
    rule_string = """
    getSubjectArmCode=="Arm A: ONC-392 + LU 177 VIPIVOTIDE" || getSubjectArmCode=="Arm 1: ONC-392 LOW DOSE + LU 177 VIPIVOTIDE" || getSubjectArmCode=="Arm 2: ONC-392 HIGH DOSE + LU 177 VIPIVOTIDE"
    """
    expression = EdkRule.expression(rule_string)
    data = expression.tree_data()
    EdkRule.draw(rule_string, "rule9.html")

    assert data == {"name": "Anonymous", "children": [{"name": "Anonymous", "children": [{"name": "Anonymous", "children": [{"name": "getSubjectArmCode", "value": "getSubjectArmCode", "type": 1}, {"name": "==", "value": "==", "type": 3}, {"name": "\"Arm A: ONC-392 + LU 177 VIPIVOTIDE\"", "value": "\"Arm A: ONC-392 + LU 177 VIPIVOTIDE\"", "type": 1}], "value": "getSubjectArmCode==\"Arm A: ONC-392 + LU 177 VIPIVOTIDE\"", "type": "statement"}, {"name": "||", "value": "||", "type": 21}, {"name": "Anonymous", "children": [{"name": "getSubjectArmCode", "value": "getSubjectArmCode", "type": 1}, {"name": "==", "value": "==", "type": 3}, {"name": "\"Arm 1: ONC-392 LOW DOSE + LU 177 VIPIVOTIDE\"", "value": "\"Arm 1: ONC-392 LOW DOSE + LU 177 VIPIVOTIDE\"", "type": 1}], "value": "getSubjectArmCode==\"Arm 1: ONC-392 LOW DOSE + LU 177 VIPIVOTIDE\"", "type": "statement"}], "value": "getSubjectArmCode==\"Arm A: ONC-392 + LU 177 VIPIVOTIDE\"||getSubjectArmCode==\"Arm 1: ONC-392 LOW DOSE + LU 177 VIPIVOTIDE\"", "type": "statement"}, {"name": "||", "value": "||", "type": 21}, {"name": "Anonymous", "children": [{"name": "getSubjectArmCode", "value": "getSubjectArmCode", "type": 1}, {"name": "==", "value": "==", "type": 3}, {"name": "\"Arm 2: ONC-392 HIGH DOSE + LU 177 VIPIVOTIDE\"", "value": "\"Arm 2: ONC-392 HIGH DOSE + LU 177 VIPIVOTIDE\"", "type": 1}], "value": "getSubjectArmCode==\"Arm 2: ONC-392 HIGH DOSE + LU 177 VIPIVOTIDE\"", "type": "statement"}], "value": "\n    getSubjectArmCode==\"Arm A: ONC-392 + LU 177 VIPIVOTIDE\" || getSubjectArmCode==\"Arm 1: ONC-392 LOW DOSE + LU 177 VIPIVOTIDE\" || getSubjectArmCode==\"Arm 2: ONC-392 HIGH DOSE + LU 177 VIPIVOTIDE\"\n    ", "type": "statement"}



def test_case10():
    rule_string = """toDate($C1D1.ONC-392 Administration.ECSTDAT)!=""&& toDate($*.*.LBDAT)!=""&& toDate("2015-1-12 "+$C1D1.ONC-392 Administration.ECSTTIM+":00")!="" &&toDate("2015-1-12 "+$*.*.*+":00")!=""?dateDiff($C1D1.ONC-392 Administration.ECSTDAT+" "+$C1D1.ONC-392 Administration.ECSTTIM+":00",$*.*.LBDAT+" "+$*.*.*+":00","m")>0:true"""
    expression = EdkRule.expression(rule_string)
    data = expression.tree_data()
    EdkRule.draw(rule_string, "rule10.html")
    out_put(data)
    assert data == {"name": "TernaryOperator", "children": [{"name": "Anonymous", "children": [{"name": "Anonymous", "children": [{"name": "toDate", "children": [{"name": "$C1D1.ONC-392 Administration.ECSTDAT", "value": "$C1D1.ONC-392 Administration.ECSTDAT", "type": 9}], "value": "toDate($C1D1.ONC-392 Administration.ECSTDAT)", "type": "statement"}, {"name": "!=", "value": "!=", "type": 4}, {"name": "\"\"", "value": "\"\"", "type": 1}], "value": "toDate($C1D1.ONC-392 Administration.ECSTDAT)!=\"\"", "type": "statement"}, {"name": "&&", "value": "&&", "type": 19}, {"name": "Anonymous", "children": [{"name": "toDate", "children": [{"name": "$*.*.LBDAT", "value": "$*.*.LBDAT", "type": 9}], "value": "toDate($*.*.LBDAT)", "type": "statement"}, {"name": "!=", "value": "!=", "type": 4}, {"name": "\"\"", "value": "\"\"", "type": 1}], "value": "toDate($*.*.LBDAT)!=\"\"", "type": "statement"}, {"name": "&&", "value": "&&", "type": 19}, {"name": "Anonymous", "children": [{"name": "toDate", "children": [{"name": "Anonymous", "children": [{"name": "Anonymous", "children": [{"name": "\"2015-1-12 \"", "value": "\"2015-1-12 \"", "type": 1}, {"name": "+", "value": "+", "type": 15}, {"name": "$C1D1.ONC-392 Administration.ECSTTIM", "value": "$C1D1.ONC-392 Administration.ECSTTIM", "type": 9}], "value": "\"2015-1-12 \"+$C1D1.ONC-392 Administration.ECSTTIM", "type": "statement"}, {"name": "+", "value": "+", "type": 15}, {"name": "\":00\"", "value": "\":00\"", "type": 1}], "value": "\"2015-1-12 \"+$C1D1.ONC-392 Administration.ECSTTIM+\":00\"", "type": "statement"}], "value": "toDate(\"2015-1-12 \"+$C1D1.ONC-392 Administration.ECSTTIM+\":00\")", "type": "statement"}, {"name": "!=", "value": "!=", "type": 4}, {"name": "\"\"", "value": "\"\"", "type": 1}], "value": "toDate(\"2015-1-12 \"+$C1D1.ONC-392 Administration.ECSTTIM+\":00\")!=\"\"", "type": "statement"}, {"name": "&&", "value": "&&", "type": 19}, {"name": "Anonymous", "children": [{"name": "toDate", "children": [{"name": "Anonymous", "children": [{"name": "Anonymous", "children": [{"name": "\"2015-1-12 \"", "value": "\"2015-1-12 \"", "type": 1}, {"name": "+", "value": "+", "type": 15}, {"name": "$*.*.*", "value": "$*.*.*", "type": 9}], "value": "\"2015-1-12 \"+$*.*.*", "type": "statement"}, {"name": "+", "value": "+", "type": 15}, {"name": "\":00\"", "value": "\":00\"", "type": 1}], "value": "\"2015-1-12 \"+$*.*.*+\":00\"", "type": "statement"}], "value": "toDate(\"2015-1-12 \"+$*.*.*+\":00\")", "type": "statement"}, {"name": "!=", "value": "!=", "type": 4}, {"name": "\"\"", "value": "\"\"", "type": 1}], "value": "toDate(\"2015-1-12 \"+$*.*.*+\":00\")!=\"\"", "type": "statement"}], "value": "toDate($C1D1.ONC-392 Administration.ECSTDAT)!=\"\"&&toDate($*.*.LBDAT)!=\"\"&&toDate(\"2015-1-12 \"+$C1D1.ONC-392 Administration.ECSTTIM+\":00\")!=\"\"&&toDate(\"2015-1-12 \"+$*.*.*+\":00\")!=\"\"", "type": "statement"}, {"name": "?", "value": "?", "type": 30}, {"name": "Anonymous", "children": [{"name": "dateDiff", "children": [{"name": "Anonymous", "children": [{"name": "Anonymous", "children": [{"name": "Anonymous", "children": [{"name": "$C1D1.ONC-392 Administration.ECSTDAT", "value": "$C1D1.ONC-392 Administration.ECSTDAT", "type": 9}, {"name": "+", "value": "+", "type": 15}, {"name": "\" \"", "value": "\" \"", "type": 1}], "value": "$C1D1.ONC-392 Administration.ECSTDAT+\" \"", "type": "statement"}, {"name": "+", "value": "+", "type": 15}, {"name": "$C1D1.ONC-392 Administration.ECSTTIM", "value": "$C1D1.ONC-392 Administration.ECSTTIM", "type": 9}], "value": "$C1D1.ONC-392 Administration.ECSTDAT+\" \"+$C1D1.ONC-392 Administration.ECSTTIM", "type": "statement"}, {"name": "+", "value": "+", "type": 15}, {"name": "\":00\"", "value": "\":00\"", "type": 1}], "value": "$C1D1.ONC-392 Administration.ECSTDAT+\" \"+$C1D1.ONC-392 Administration.ECSTTIM+\":00\"", "type": "statement"}, {"name": "Anonymous", "children": [{"name": "Anonymous", "children": [{"name": "Anonymous", "children": [{"name": "$*.*.LBDAT", "value": "$*.*.LBDAT", "type": 9}, {"name": "+", "value": "+", "type": 15}, {"name": "\" \"", "value": "\" \"", "type": 1}], "value": "$*.*.LBDAT+\" \"", "type": "statement"}, {"name": "+", "value": "+", "type": 15}, {"name": "$*.*.*", "value": "$*.*.*", "type": 9}], "value": "$*.*.LBDAT+\" \"+$*.*.*", "type": "statement"}, {"name": "+", "value": "+", "type": 15}, {"name": "\":00\"", "value": "\":00\"", "type": 1}], "value": "$*.*.LBDAT+\" \"+$*.*.*+\":00\"", "type": "statement"}, {"name": "\"m\"", "value": "\"m\"", "type": 1}], "value": "dateDiff($C1D1.ONC-392 Administration.ECSTDAT+\" \"+$C1D1.ONC-392 Administration.ECSTTIM+\":00\",$*.*.LBDAT+\" \"+$*.*.*+\":00\",\"m\")", "type": "statement"}, {"name": ">", "value": ">", "type": 7}, {"name": "0", "value": "0", "type": 17}], "value": "dateDiff($C1D1.ONC-392 Administration.ECSTDAT+\" \"+$C1D1.ONC-392 Administration.ECSTTIM+\":00\",$*.*.LBDAT+\" \"+$*.*.*+\":00\",\"m\")>0", "type": "statement"}, {"name": ":", "value": ":", "type": 14}, {"name": "true", "value": "true", "type": 22}], "value": "toDate($C1D1.ONC-392 Administration.ECSTDAT)!=\"\"&& toDate($*.*.LBDAT)!=\"\"&& toDate(\"2015-1-12 \"+$C1D1.ONC-392 Administration.ECSTTIM+\":00\")!=\"\" &&toDate(\"2015-1-12 \"+$*.*.*+\":00\")!=\"\"?dateDiff($C1D1.ONC-392 Administration.ECSTDAT+\" \"+$C1D1.ONC-392 Administration.ECSTTIM+\":00\",$*.*.LBDAT+\" \"+$*.*.*+\":00\",\"m\")>0:true", "type": "statement"}


def test_case11():
    rule_string = """autoValue(RoundN(sum(getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLLONG, $*.Target Lesions Assessment (Details) (Screening).TLLOC != 11 && $*.Target Lesions Assessment (Details) (Screening).TLDIAU == "CM") , multiply(getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLLONG, $*.Target Lesions Assessment (Details) (Screening).TLLOC != 11 && $*.Target Lesions Assessment (Details) (Screening).TLDIAU == "MM"),1/10) , getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLSHORT, $*.Target Lesions Assessment (Details) (Screening).TLLOC == 11 && $*.Target Lesions Assessment (Details) (Screening).TLDIAU== "CM") , multiply(getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLSHORT, $*.Target Lesions Assessment (Details) (Screening).TLLOC == 11 && $*.Target Lesions Assessment (Details) (Screening).TLDIAU== "MM"),1/10)), 0.01), true)"""
    expression = EdkRule.expression(rule_string)
    data = expression.tree_data()
    EdkRule.draw(rule_string, "rule11.html")
    out_put(data)
    assert data == {"name": "autoValue", "children": [{"name": "RoundN", "children": [{"name": "sum", "children": [{"name": "getSumOfItemInLog", "children": [{"name": "$*.Target Lesions Assessment (Details) (Screening).TLLONG", "value": "$*.Target Lesions Assessment (Details) (Screening).TLLONG", "type": 9}, {"name": "Anonymous", "children": [{"name": "Anonymous", "children": [{"name": "$*.Target Lesions Assessment (Details) (Screening).TLLOC ", "value": "$*.Target Lesions Assessment (Details) (Screening).TLLOC ", "type": 9}, {"name": "!=", "value": "!=", "type": 4}, {"name": "11", "value": "11", "type": 17}], "value": "$*.Target Lesions Assessment (Details) (Screening).TLLOC !=11", "type": "statement"}, {"name": "&&", "value": "&&", "type": 19}, {"name": "Anonymous", "children": [{"name": "$*.Target Lesions Assessment (Details) (Screening).TLDIAU ", "value": "$*.Target Lesions Assessment (Details) (Screening).TLDIAU ", "type": 9}, {"name": "==", "value": "==", "type": 3}, {"name": "\"CM\"", "value": "\"CM\"", "type": 1}], "value": "$*.Target Lesions Assessment (Details) (Screening).TLDIAU ==\"CM\"", "type": "statement"}], "value": "$*.Target Lesions Assessment (Details) (Screening).TLLOC !=11&&$*.Target Lesions Assessment (Details) (Screening).TLDIAU ==\"CM\"", "type": "statement"}], "value": "getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLLONG,$*.Target Lesions Assessment (Details) (Screening).TLLOC !=11&&$*.Target Lesions Assessment (Details) (Screening).TLDIAU ==\"CM\")", "type": "statement"}, {"name": "multiply", "children": [{"name": "getSumOfItemInLog", "children": [{"name": "$*.Target Lesions Assessment (Details) (Screening).TLLONG", "value": "$*.Target Lesions Assessment (Details) (Screening).TLLONG", "type": 9}, {"name": "Anonymous", "children": [{"name": "Anonymous", "children": [{"name": "$*.Target Lesions Assessment (Details) (Screening).TLLOC ", "value": "$*.Target Lesions Assessment (Details) (Screening).TLLOC ", "type": 9}, {"name": "!=", "value": "!=", "type": 4}, {"name": "11", "value": "11", "type": 17}], "value": "$*.Target Lesions Assessment (Details) (Screening).TLLOC !=11", "type": "statement"}, {"name": "&&", "value": "&&", "type": 19}, {"name": "Anonymous", "children": [{"name": "$*.Target Lesions Assessment (Details) (Screening).TLDIAU ", "value": "$*.Target Lesions Assessment (Details) (Screening).TLDIAU ", "type": 9}, {"name": "==", "value": "==", "type": 3}, {"name": "\"MM\"", "value": "\"MM\"", "type": 1}], "value": "$*.Target Lesions Assessment (Details) (Screening).TLDIAU ==\"MM\"", "type": "statement"}], "value": "$*.Target Lesions Assessment (Details) (Screening).TLLOC !=11&&$*.Target Lesions Assessment (Details) (Screening).TLDIAU ==\"MM\"", "type": "statement"}], "value": "getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLLONG,$*.Target Lesions Assessment (Details) (Screening).TLLOC !=11&&$*.Target Lesions Assessment (Details) (Screening).TLDIAU ==\"MM\")", "type": "statement"}, {"name": "Anonymous", "children": [{"name": "1", "value": "1", "type": 17}, {"name": "/", "value": "/", "type": 36}, {"name": "10", "value": "10", "type": 17}], "value": "1/10", "type": "statement"}], "value": "multiply(getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLLONG,$*.Target Lesions Assessment (Details) (Screening).TLLOC !=11&&$*.Target Lesions Assessment (Details) (Screening).TLDIAU ==\"MM\"),1/10)", "type": "statement"}, {"name": "getSumOfItemInLog", "children": [{"name": "$*.Target Lesions Assessment (Details) (Screening).TLSHORT", "value": "$*.Target Lesions Assessment (Details) (Screening).TLSHORT", "type": 9}, {"name": "Anonymous", "children": [{"name": "Anonymous", "children": [{"name": "$*.Target Lesions Assessment (Details) (Screening).TLLOC ", "value": "$*.Target Lesions Assessment (Details) (Screening).TLLOC ", "type": 9}, {"name": "==", "value": "==", "type": 3}, {"name": "11", "value": "11", "type": 17}], "value": "$*.Target Lesions Assessment (Details) (Screening).TLLOC ==11", "type": "statement"}, {"name": "&&", "value": "&&", "type": 19}, {"name": "Anonymous", "children": [{"name": "$*.Target Lesions Assessment (Details) (Screening).TLDIAU", "value": "$*.Target Lesions Assessment (Details) (Screening).TLDIAU", "type": 9}, {"name": "==", "value": "==", "type": 3}, {"name": "\"CM\"", "value": "\"CM\"", "type": 1}], "value": "$*.Target Lesions Assessment (Details) (Screening).TLDIAU==\"CM\"", "type": "statement"}], "value": "$*.Target Lesions Assessment (Details) (Screening).TLLOC ==11&&$*.Target Lesions Assessment (Details) (Screening).TLDIAU==\"CM\"", "type": "statement"}], "value": "getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLSHORT,$*.Target Lesions Assessment (Details) (Screening).TLLOC ==11&&$*.Target Lesions Assessment (Details) (Screening).TLDIAU==\"CM\")", "type": "statement"}, {"name": "multiply", "children": [{"name": "getSumOfItemInLog", "children": [{"name": "$*.Target Lesions Assessment (Details) (Screening).TLSHORT", "value": "$*.Target Lesions Assessment (Details) (Screening).TLSHORT", "type": 9}, {"name": "Anonymous", "children": [{"name": "Anonymous", "children": [{"name": "$*.Target Lesions Assessment (Details) (Screening).TLLOC ", "value": "$*.Target Lesions Assessment (Details) (Screening).TLLOC ", "type": 9}, {"name": "==", "value": "==", "type": 3}, {"name": "11", "value": "11", "type": 17}], "value": "$*.Target Lesions Assessment (Details) (Screening).TLLOC ==11", "type": "statement"}, {"name": "&&", "value": "&&", "type": 19}, {"name": "Anonymous", "children": [{"name": "$*.Target Lesions Assessment (Details) (Screening).TLDIAU", "value": "$*.Target Lesions Assessment (Details) (Screening).TLDIAU", "type": 9}, {"name": "==", "value": "==", "type": 3}, {"name": "\"MM\"", "value": "\"MM\"", "type": 1}], "value": "$*.Target Lesions Assessment (Details) (Screening).TLDIAU==\"MM\"", "type": "statement"}], "value": "$*.Target Lesions Assessment (Details) (Screening).TLLOC ==11&&$*.Target Lesions Assessment (Details) (Screening).TLDIAU==\"MM\"", "type": "statement"}], "value": "getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLSHORT,$*.Target Lesions Assessment (Details) (Screening).TLLOC ==11&&$*.Target Lesions Assessment (Details) (Screening).TLDIAU==\"MM\")", "type": "statement"}, {"name": "Anonymous", "children": [{"name": "1", "value": "1", "type": 17}, {"name": "/", "value": "/", "type": 36}, {"name": "10", "value": "10", "type": 17}], "value": "1/10", "type": "statement"}], "value": "multiply(getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLSHORT,$*.Target Lesions Assessment (Details) (Screening).TLLOC ==11&&$*.Target Lesions Assessment (Details) (Screening).TLDIAU==\"MM\"),1/10)", "type": "statement"}], "value": "sum(getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLLONG,$*.Target Lesions Assessment (Details) (Screening).TLLOC !=11&&$*.Target Lesions Assessment (Details) (Screening).TLDIAU ==\"CM\"),multiply(getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLLONG,$*.Target Lesions Assessment (Details) (Screening).TLLOC !=11&&$*.Target Lesions Assessment (Details) (Screening).TLDIAU ==\"MM\"),1/10),getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLSHORT,$*.Target Lesions Assessment (Details) (Screening).TLLOC ==11&&$*.Target Lesions Assessment (Details) (Screening).TLDIAU==\"CM\"),multiply(getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLSHORT,$*.Target Lesions Assessment (Details) (Screening).TLLOC ==11&&$*.Target Lesions Assessment (Details) (Screening).TLDIAU==\"MM\"),1/10))", "type": "statement"}, {"name": "0.01", "value": "0.01", "type": 17}], "value": "RoundN(sum(getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLLONG,$*.Target Lesions Assessment (Details) (Screening).TLLOC !=11&&$*.Target Lesions Assessment (Details) (Screening).TLDIAU ==\"CM\"),multiply(getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLLONG,$*.Target Lesions Assessment (Details) (Screening).TLLOC !=11&&$*.Target Lesions Assessment (Details) (Screening).TLDIAU ==\"MM\"),1/10),getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLSHORT,$*.Target Lesions Assessment (Details) (Screening).TLLOC ==11&&$*.Target Lesions Assessment (Details) (Screening).TLDIAU==\"CM\"),multiply(getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLSHORT,$*.Target Lesions Assessment (Details) (Screening).TLLOC ==11&&$*.Target Lesions Assessment (Details) (Screening).TLDIAU==\"MM\"),1/10)),0.01)", "type": "statement"}, {"name": "true", "value": "true", "type": 22}], "value": "autoValue(RoundN(sum(getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLLONG, $*.Target Lesions Assessment (Details) (Screening).TLLOC != 11 && $*.Target Lesions Assessment (Details) (Screening).TLDIAU == \"CM\") , multiply(getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLLONG, $*.Target Lesions Assessment (Details) (Screening).TLLOC != 11 && $*.Target Lesions Assessment (Details) (Screening).TLDIAU == \"MM\"),1/10) , getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLSHORT, $*.Target Lesions Assessment (Details) (Screening).TLLOC == 11 && $*.Target Lesions Assessment (Details) (Screening).TLDIAU== \"CM\") , multiply(getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLSHORT, $*.Target Lesions Assessment (Details) (Screening).TLLOC == 11 && $*.Target Lesions Assessment (Details) (Screening).TLDIAU== \"MM\"),1/10)), 0.01), true)", "type": "statement"}


def test_case12():
    rule_string = """mustAnswer($*.*.CRONGO)&&$*.*.CRONGO=='N'?$*.*.*!='':true"""
    expression = EdkRule.expression(rule_string)
    data = expression.tree_data()
    EdkRule.draw(rule_string, "rule12.html")

    assert data == {"name": "TernaryOperator", "children": [{"name": "Anonymous", "children": [{"name": "mustAnswer", "children": [{"name": "$*.*.CRONGO", "value": "$*.*.CRONGO", "type": 9}], "value": "mustAnswer($*.*.CRONGO)", "type": "statement"}, {"name": "&&", "value": "&&", "type": 19}, {"name": "Anonymous", "children": [{"name": "$*.*.CRONGO", "value": "$*.*.CRONGO", "type": 9}, {"name": "==", "value": "==", "type": 3}, {"name": "'N'", "value": "'N'", "type": 1}], "value": "$*.*.CRONGO=='N'", "type": "statement"}], "value": "mustAnswer($*.*.CRONGO)&&$*.*.CRONGO=='N'", "type": "statement"}, {"name": "?", "value": "?", "type": 30}, {"name": "Anonymous", "children": [{"name": "$*.*.*", "value": "$*.*.*", "type": 9}, {"name": "!=", "value": "!=", "type": 4}, {"name": "''", "value": "''", "type": 1}], "value": "$*.*.*!=''", "type": "statement"}, {"name": ":", "value": ":", "type": 14}, {"name": "true", "value": "true", "type": 22}], "value": "mustAnswer($*.*.CRONGO)&&$*.*.CRONGO=='N'?$*.*.*!='':true", "type": "statement"}


def test_case13():
    rule_string = """a==1&&b==2||c"""
    expression = EdkRule.expression(rule_string)
    data = expression.tree_data()
    EdkRule.draw(rule_string, "rule13.html")
    assert data == {"name": "Anonymous", "children": [{"name": "Anonymous", "children": [{"name": "Anonymous", "children": [{"name": "a", "value": "a", "type": 1}, {"name": "==", "value": "==", "type": 3}, {"name": "1", "value": "1", "type": 17}], "value": "a==1", "type": "statement"}, {"name": "&&", "value": "&&", "type": 19}, {"name": "Anonymous", "children": [{"name": "b", "value": "b", "type": 1}, {"name": "==", "value": "==", "type": 3}, {"name": "2", "value": "2", "type": 17}], "value": "b==2", "type": "statement"}], "value": "a==1&&b==2", "type": "statement"}, {"name": "||", "value": "||", "type": 21}, {"name": "c", "value": "c", "type": 1}], "value": "a==1&&b==2||c", "type": "statement"}



def test_rule_draw():
    test_case1()
    test_case2()
    test_case3()
    test_case4()
    test_case5()
    test_case6()
    test_case7()
    test_case8()
    test_case9()
    test_case10()
    test_case11()
    test_case12()
    test_case13()









from edkrule.interpreter.lexer.character import Character
from edkrule.interpreter.lexer.dfa_state import DfaState
from edkrule.interpreter.lexer.token_type import TokenType


class GeToken:
    @staticmethod
    def accept(lexer, char: str):
        if lexer.state == DfaState.Gt:
            if Character.isassign(char):
                lexer.state = DfaState.Ge
                lexer.token_type = TokenType.Ge
                lexer.token_text += char
                return True
        return False



from edkrule.interpreter.lexer.dfa_state import DfaState
from edkrule.interpreter.lexer.token_type import TokenType


class EqMinusState:
    @staticmethod
    def accept(lexer, char: str, state: DfaState):
        if state == DfaState.EqMinus:
            lexer.token_type = TokenType.EqMinus
            lexer.init_token(char)
            return True
        return False



from edkrule.interpreter.lexer.character import Character
from edkrule.interpreter.lexer.dfa_state import DfaState
from edkrule.interpreter.lexer.token_type import TokenType


class BiteOrState:
    @staticmethod
    def accept(lexer, char: str, state: DfaState):
        if state == DfaState.ByteOr and Character.isL(char):
            lexer.token_text += char
            lexer.state = DfaState.Or
            return True
        elif state == DfaState.ByteOr:
            if Character.isseparators(char) or Character.iskeyword(char):
                lexer.token_type = TokenType.ByteOr
                lexer.init_token(char)
            else:
                lexer.token_text += char
                lexer.state = DfaState.Identifier
            return True
        return False


from edkrule.interpreter.lexer.token_type import TokenType


class Categorize:
    @staticmethod
    def is_op(stmt):
        return stmt.type in [TokenType.Plus, TokenType.Minus, TokenType.DPlus, TokenType.DMinus,
                             TokenType.Multipy, TokenType.Divide, TokenType.And, TokenType.Or,
                             TokenType.Eq, TokenType.Ep, TokenType.NEq, TokenType.Nidentity,
                             TokenType.EqPlus, TokenType.MinusEq, TokenType.EqMinus, TokenType.PlusEq,
                             TokenType.Assignment, TokenType.Colon, TokenType.Question,
                             TokenType.Gt, TokenType.Ge, TokenType.Le, TokenType.Lt]

    @staticmethod
    def is_variable(stmt):
        return stmt.type in [TokenType.Identifier, TokenType.Lp,
                             TokenType.RealNumber, TokenType.TRUE,
                             TokenType.FALSE, TokenType.Variable]

    @staticmethod
    def is_exp(stmt):
        return stmt.type in [TokenType.Identifier, TokenType.Lp,
                             TokenType.RealNumber, TokenType.TRUE,
                             TokenType.FALSE, TokenType.Expression,
                             TokenType.Variable]

    @staticmethod
    def is_stmt(stmt):
        return stmt.type == TokenType.Statement


from edkrule.engine.finder.by import By


class Result:
    def __init__(self, expression):
        self._expression = expression

    def get(self, rule_fragment_text: str = None):
        """
        返回表达式的执行结果
        如果 rule_fragment_text 不传，则返回当前Expression 的执行结果
        如果 rule_fragment_text 传入片段，则返回片段的执行结果
        :param rule_fragment_text:
        :type rule_fragment_text:
        :return:
        :rtype:
        """
        if not rule_fragment_text:
            return self._expression.engine.cache.get(self._expression.rid)
        else:
            rid = self._expression.find(rule_fragment_text).rid
            return self._expression.engine.cache.get(rid)

    def track(self):
        """返回执行顺序及对应的结果"""
        return [{
            rid: {"text": self._expression.find(By.Rid, rid).text,
                  "result": self._expression.engine.cache.get(rid)
                  }
        } for rid in self._expression.engine.squeue.data]



from edkrule.interpreter.expression.stack import Stack
from edkrule.interpreter.lexer.token_type import TokenType
from edkrule.interpreter.parse.statement import Statement
from edkrule.interpreter.parse.statement_parser.base_statement_parser import AbcStatementParser
from edkrule.interpreter.parse.ternary_stack import TernaryStack


class TernaryStatementParser(AbcStatementParser):
    def __init__(self, parser):
        super().__init__(parser)
        self.left_statement = Statement()
        self.right_statement = Statement()
        self.left_flag = TernaryStack()
        self.right_flag = TernaryStack()
        self.left_content = []
        self.right_content = []
        self.bracket_stack = Stack()

    def accept(self) -> bool:

        if self.parser.index <= len(self.parser.body) - 1:
            if self.token().type == TokenType.Question:
                self.parser.statement.add(self.token())
                self.left_flag.append(self.token())
                self.left()
                self.right()
                return True
            # self.move()
        return False

    def ternary_end(self):
        if self.parser.index > len(self.parser.body) - 1:
            return True
        if self.token().type == TokenType.Comma:
            if self.bracket_stack.empty():
                return True
        if self.token().type == TokenType.Rp:
            if self.bracket_stack.empty():
                return True
        return False
        # return self.parser.index > len(
        #     self.parser.content) - 1 or self.token().type == TokenType.Comma or self.token().type == TokenType.Rp

    def left(self):
        from edkrule.interpreter.parse.parser import Parser
        while not self.left_flag.empty():
            self.move()

            if self.token().type == TokenType.Question:
                self.left_flag.append(self.token())
                self.left_content.append(self.token())
            elif self.token().type == TokenType.Colon:
                self.left_flag.pop()
                if not self.left_flag.empty(): self.left_content.append(self.token())
            else:
                self.left_content.append(self.token())
        parser = Parser(self.left_content).parse()
        self.left_statement = parser.statement
        self.right_flag.append(self.token())
        self.parser.statement.add(self.left_statement)
        self.parser.statement.add(self.token())

    def right(self):
        from edkrule.interpreter.parse.parser import Parser
        while not self.right_flag.empty():
            self.move()
            if self.parser.index <= len(self.parser.body) - 1:
                if self.token().type == TokenType.Lp:
                    self.bracket_stack.push(self.token())
                if self.token().type == TokenType.Rp:
                    self.bracket_stack.pop()
            # self.right_content.append(self.token())
            if self.ternary_end():
                if self.parser.index <= len(self.parser.body) - 1:
                    if self.token().type == TokenType.Rp:
                        self.right_content.append(self.token())
                self.right_flag.pop()
                parser = Parser(self.right_content).parse()
                self.right_statement = parser.statement
                # self.right_flag.append(self.token())
                self.parser.statement.add(self.right_statement)
            else:
                self.right_content.append(self.token())



from edkrule.interpreter.lexer.character import Character
from edkrule.interpreter.lexer.dfa_state import DfaState
from edkrule.interpreter.lexer.token_type import TokenType


class TrueState:

    @staticmethod
    def accept(lexer, char: str, state: DfaState):
        if state == DfaState.True1:
            lexer.token_text += char
            lexer.state = DfaState.True2 if Character.isr(char) else DfaState.Identifier
            # lexer.state = DfaState.True2
            return True
        elif state == DfaState.True2:
            lexer.token_text += char
            lexer.state = DfaState.True3 if Character.isu(char) else DfaState.Identifier
            # lexer.state = DfaState.True3
            return True
        elif state == DfaState.True3:
            lexer.token_text += char
            lexer.state = DfaState.True4 if Character.ise(char) else DfaState.Identifier
            # lexer.state = DfaState.True4
            return True
        elif state == DfaState.True4:
            if Character.isseparators(char) or Character.iscolon(char):
                lexer.token_type = TokenType.TRUE
                lexer.init_token(char)
            else:
                lexer.token_text += char
                lexer.state = DfaState.Identifier
            return True
        return False





from io import StringIO

from edkrule.engine.finder.finder import Finder
from edkrule.interpreter.lexer.lexer import Lexer
from edkrule.interpreter.lexer.token_type import TokenType
from edkrule.interpreter.lexer.tokens.token import Token


class FinderByRuleFragment(Finder):
    def __init__(self, formate=None):
        self._format = formate

    def find(self, expression, rule_fragment_text):
        if self._format:
            rule_fragment_text = self._format.initialize(rule_fragment_text)
        if rule_fragment_text == expression.text:
            return expression
        else:
            for body in expression.body:
                if type(body) != Token:
                    rule = body.find(FinderByRuleFragment(self._format), rule_fragment_text)
                    if rule is not None: return rule

    # def format(self, rule_fragment_text):
    #     lexer = Lexer()
    #     lexer.tokenize(StringIO(rule_fragment_text))
    #     text = "".join([e.text for e in list(filter(lambda e: e.type != TokenType.Blank, l.token_list))])
    #     return text



import uuid

from edkrule.engine.constant import Constant
from edkrule.engine.engine import Engine
from edkrule.engine.finder.finder import Finder

from edkrule.interpreter.lexer.token_type import TokenType


class Statement:
    def __init__(self):
        self.body = []
        self._engine = None
        self.rid = uuid.uuid4().hex

    def add(self, e):
        self.body.append(e)

    @property
    def engine(self) -> Engine:
        return self._engine

    @engine.setter
    def engine(self, value):
        self._engine = value

    def run(self):
        runner = None
        if self.type == TokenType.Statement:
            raise "Statement can not run"
        if self.type == TokenType.RealNumber:
            runner = self.engine.get_class(Constant.DataTypes, TokenType.RealNumber.name)(self.engine, self.rid)
        elif self.type == TokenType.Identifier:
            runner = self.engine.get_class(Constant.Identifiers, self.text)(self.engine, self.rid)
        elif self.type == TokenType.TRUE or self.type == TokenType.FALSE:
            runner = self.engine.get_class(Constant.DataTypes, self.text)(self.engine, self.rid)
        elif self.type == TokenType.Variable:
            runner = self.engine.get_class(Constant.DataTypes, TokenType.Variable.name)(self.engine, self.rid)
        result = runner.run(self.text)
        return result

    @property
    def type(self):
        if len(self.body) == 1: return self.body[0].type
        return TokenType.Statement

    def token(self):
        if len(self.body) == 1: return self.body[0]
        raise Exception("Only Statement contain one element support")

    @property
    def text(self):
        if len(self.body) == 1: return self.body[0].text
        return "".join([e.text for e in self.body])

    def empty(self):
        return len(self.body) == 0

    def count(self):
        return len(self.body)

    def remove(self, from_i: int):
        self.body = self.body[0:from_i + 1]

    def matrix(self, *args):
        deep = len(args)
        index = 0
        result = self.body[args[index]]
        while deep > 1:
            deep -= 1
            index += 1
            result = result.body[args[index]]

        return result

    def display(self, index: list):
        """
        测试用
        :param index:
        :type index:
        :return:
        :rtype:
        """
        if not index: print(self.text)
        for i, e in enumerate(self.body):
            pos = []
            pos.extend(index)
            pos.append(str(i))
            print("assert", f'es.statement.matrix({",".join(pos)}).text == \'{e.text}\'')
            print("assert", f'es.statement.matrix({",".join(pos)}).type == {e.type}')
            # if Categorize.is_variable(e) or Categorize.is_op(e):
            #     # exp.matrix(0).name == 'Anonymous'
            #
            # else:
            #     print("assert", f'exp.matrix({",".join(pos)}).name == \'{e.name}\'')
            #     print("assert", f'exp.matrix({",".join(pos)}).type == {e.type}')
            #     # print(",".join(pos), e.name, e.type)
            if e.type == TokenType.Statement:
                e.display(pos)

    def find(self, finder: Finder, finder_parameter):
        return finder.find(self, finder_parameter)




from edkrule.interpreter.lexer.character import Character
from edkrule.interpreter.lexer.dfa_state import DfaState


class DqToken:
    @staticmethod
    def accept(lexer, char: str):
        if Character.isdoublequotation(char):
            lexer.token_text += char
            lexer.state = DfaState.Dq
            return True
        return False


from edkrule.interpreter.lexer.token_type import TokenType
from edkrule.interpreter.lexer.tokens.token import Token


class Priority:
    @staticmethod
    def level(token: Token):
        if token.type in [TokenType.Plus, TokenType.Minus, TokenType.DPlus, TokenType.DMinus]: return 5
        if token.type in [TokenType.Multipy, TokenType.Divide]: return 6
        # if token.type in [TokenType.And, TokenType.Or, TokenType.Eq,
        if token.type in [TokenType.Eq,
                          TokenType.Ep, TokenType.NEq, TokenType.Nidentity,
                          TokenType.EqPlus, TokenType.MinusEq, TokenType.EqMinus,
                          TokenType.PlusEq, TokenType.Gt, TokenType.Ge, TokenType.Lt, TokenType.Le]: return 0
        if token.type in [TokenType.And, TokenType.Or]: return -1
        if token.type in [TokenType.Colon, TokenType.Question]: return -2
        if token.type in [TokenType.Assignment]: return -10

    @staticmethod
    def compare(token: Token, with_token: Token):
        if Priority.level(token) > Priority.level(with_token):
            return 1
        elif Priority.level(token) == Priority.level(with_token):
            return 0
        else:
            return -1

    @staticmethod
    def gt(token: Token, with_token: Token):
        return Priority.compare(token, with_token) == 1

    @staticmethod
    def ge(token: Token, with_token: Token):
        return Priority.compare(token, with_token) == 1 or Priority.compare(token, with_token) == 0

    @staticmethod
    def lt(token: Token, with_token: Token):
        return Priority.compare(token, with_token) == -1

    @staticmethod
    def eq(token: Token, with_token: Token):
        return Priority.compare(token, with_token) == 0


import abc

from edkrule.utils.cache import Cache
from edkrule.utils.squeue import Squeue


class Runner(abc.ABC):
    def __init__(self, engine, rid):
        self._engine = engine
        self.rid = rid

    def run(self, *args):
        """
        step1. 计算结果
        step2. 计算结果存储在 cache 中
        :param args:
        :type args:
        :return:
        :rtype:
        """
        result = self.execute(*args)
        self.cache().set(self.rid, result)
        self.squeue().append(self.rid)
        return result

    @abc.abstractmethod
    def execute(self, *args): ...

    def cache(self) -> Cache:
        return self._engine.cache

    def squeue(self) -> Squeue:
        return self._engine.squeue



from edkrule.interpreter.lexer.token_type import TokenType
from edkrule.interpreter.parse.statement import Statement
from edkrule.interpreter.parse.statement_parser.base_statement_parser import AbcStatementParser


class QuestionStatementParser(AbcStatementParser):

    def __init__(self, parser):
        super().__init__(parser)
        self.statement = Statement()

    def accept(self):
        if self.parser.index <= len(self.parser.body) - 1:
            if self.token().type == TokenType.Question:
                self.parser.ternary_stack.push(self.token())
                self.parser.statement.add(self.token())
                self.move()
                return True
            # self.move()
        return False



from edkrule.interpreter.lexer.dfa_state import DfaState


class InitialState:

    @staticmethod
    def accept(lexer, char: str, state: DfaState):
        if state == DfaState.Initial:
            return lexer.init_token(char)



import dataclasses

from edkrule.interpreter.expression.bracket_enum import BracketEnum
from edkrule.interpreter.lexer.tokens.token import Token


@dataclasses.dataclass
class Bracket:
    type: BracketEnum
    token: Token



from io import StringIO

import pytest

from edkrule.interpreter.lexer.lexer import Lexer
from edkrule.interpreter.lexer.token_type import TokenType


@pytest.fixture
def rules():
    return [
        """toDate($*.*.*)!=""&&toDate(getMinByLog($*.Informed Consent.DSSTDAT))!=""?dateDiff(getMinByLog($*.Informed Consent.DSSTDAT),$*.*.*, "D")>0:true""",
        """isValidDate($*.*.*)""",
        """autoValue((toNum($*.*.EGORRES.881)!=''&&toNum($*.*.EGORRES.879)!=''?RoundN(toNum($*.*.EGORRES.879)/Math.pow((toNum($*.*.EGORRES.881)/1000), 0.33), "1"):""), true)""",
        "$*.*.EGPERF==\"Y\"",
        "$*.*.EGORRES.12394==\"CS\"",
        """condition(mustAnswer(toNum($*.*.*)), isRange($*.*.*, 40,160))""",
        """mustAnswer($*.*.CRONGO)&&$*.*.CRONGO=='Y'?$*.*.*=='':true""",
        """mustAnswer($*.*.CRONGO)&&$*.*.CRONGO=='N'?$*.*.*!='':true""",
        """autoIncrease(1,1,1,1)""",
        """$*.*.*=='Y'&&mustAnswer($*.Demographics.AGE)?toNum($*.Demographics.AGE)>=18:true""",
        """mustAnswer($*.*.*)&&(getICFVersion($*.Date of Visit.SVDAT, $*.Informed Consent.DSSTDAT, $*.Informed Consent.VERSION)=='V1.5'||getICFVersion($*.Date of Visit.SVDAT, $*.Informed Consent.DSSTDAT, $*.Informed Consent.VERSION)=='V1.6')?($*.*.*=='0'||$*.*.*=='1'||$*.*.*=='2'):true""",
        """toDate($C1D1.ONC-392 Administration.ECSTDAT)!=""&& toDate($*.*.LBDAT)!=""&& toDate("2015-1-12 "+$C1D1.ONC-392 Administration.ECSTTIM+":00")!="" &&toDate("2015-1-12 "+$*.*.*+":00")!=""?dateDiff($C1D1.ONC-392 Administration.ECSTDAT+" "+$C1D1.ONC-392 Administration.ECSTTIM+":00",$*.*.LBDAT+" "+$*.*.*+":00","m")>0:true""",
        """autoValue(RoundN(sum(getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLLONG, $*.Target Lesions Assessment (Details) (Screening).TLLOC != 11 && $*.Target Lesions Assessment (Details) (Screening).TLDIAU == "CM") , multiply(getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLLONG, $*.Target Lesions Assessment (Details) (Screening).TLLOC != 11 && $*.Target Lesions Assessment (Details) (Screening).TLDIAU == "MM"),1/10) , getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLSHORT, $*.Target Lesions Assessment (Details) (Screening).TLLOC == 11 && $*.Target Lesions Assessment (Details) (Screening).TLDIAU== "CM") , multiply(getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLSHORT, $*.Target Lesions Assessment (Details) (Screening).TLLOC == 11 && $*.Target Lesions Assessment (Details) (Screening).TLDIAU== "MM"),1/10)), 0.01), true)"""
    ]


def test_rule(rules):
    test_rule_lexer0(rules=rules)
    test_rule_lexer1(rules=rules)
    test_rule_lexer2(rules=rules)
    test_rule_lexer3(rules=rules)
    test_rule_lexer4(rules=rules)
    test_rule_lexer5(rules=rules)
    test_rule_lexer6(rules=rules)
    test_rule_lexer7(rules=rules)
    test_rule_lexer8(rules=rules)
    test_rule_lexer9(rules=rules)
    test_rule_lexer10(rules=rules)
    test_rule_lexer11(rules=rules)
    test_rule_lexer12(rules=rules)





def test_rule_lexer0(rules):
    string = StringIO(rules[0])
    lexer = Lexer()
    lexer.tokenize(string)
    assert lexer.token_list[0].text == 'toDate'
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[1].text == '('
    assert lexer.token_list[1].type == TokenType.Lp
    assert lexer.token_list[2].text == '$*.*.*'
    assert lexer.token_list[2].type == TokenType.Variable
    assert lexer.token_list[3].text == ')'
    assert lexer.token_list[3].type == TokenType.Rp
    assert lexer.token_list[4].text == '!='
    assert lexer.token_list[4].type == TokenType.NEq
    assert lexer.token_list[5].text == '""'
    assert lexer.token_list[5].type == TokenType.Identifier
    assert lexer.token_list[6].text == '&&'
    assert lexer.token_list[6].type == TokenType.And
    assert lexer.token_list[7].text == 'toDate'
    assert lexer.token_list[7].type == TokenType.Identifier
    assert lexer.token_list[8].text == '('
    assert lexer.token_list[8].type == TokenType.Lp
    assert lexer.token_list[9].text == 'getMinByLog'
    assert lexer.token_list[9].type == TokenType.Identifier
    assert lexer.token_list[10].text == '('
    assert lexer.token_list[10].type == TokenType.Lp
    assert lexer.token_list[11].text == '$*.Informed Consent.DSSTDAT'
    assert lexer.token_list[11].type == TokenType.Variable
    assert lexer.token_list[12].text == ')'
    assert lexer.token_list[12].type == TokenType.Rp
    assert lexer.token_list[13].text == ')'
    assert lexer.token_list[13].type == TokenType.Rp
    assert lexer.token_list[14].text == '!='
    assert lexer.token_list[14].type == TokenType.NEq
    assert lexer.token_list[15].text == '""'
    assert lexer.token_list[15].type == TokenType.Identifier
    assert lexer.token_list[16].text == '?'
    assert lexer.token_list[16].type == TokenType.Question
    assert lexer.token_list[17].text == 'dateDiff'
    assert lexer.token_list[17].type == TokenType.Identifier
    assert lexer.token_list[18].text == '('
    assert lexer.token_list[18].type == TokenType.Lp
    assert lexer.token_list[19].text == 'getMinByLog'
    assert lexer.token_list[19].type == TokenType.Identifier
    assert lexer.token_list[20].text == '('
    assert lexer.token_list[20].type == TokenType.Lp
    assert lexer.token_list[21].text == '$*.Informed Consent.DSSTDAT'
    assert lexer.token_list[21].type == TokenType.Variable
    assert lexer.token_list[22].text == ')'
    assert lexer.token_list[22].type == TokenType.Rp
    assert lexer.token_list[23].text == ','
    assert lexer.token_list[23].type == TokenType.Comma
    assert lexer.token_list[24].text == '$*.*.*'
    assert lexer.token_list[24].type == TokenType.Variable
    assert lexer.token_list[25].text == ','
    assert lexer.token_list[25].type == TokenType.Comma
    assert lexer.token_list[26].text == ' '
    assert lexer.token_list[26].type == TokenType.Blank
    assert lexer.token_list[27].text == '"D"'
    assert lexer.token_list[27].type == TokenType.Identifier
    assert lexer.token_list[28].text == ')'
    assert lexer.token_list[28].type == TokenType.Rp
    assert lexer.token_list[29].text == '>'
    assert lexer.token_list[29].type == TokenType.Gt
    assert lexer.token_list[30].text == '0'
    assert lexer.token_list[30].type == TokenType.RealNumber
    assert lexer.token_list[31].text == ':'
    assert lexer.token_list[31].type == TokenType.Colon
    assert lexer.token_list[32].text == 'true'
    assert lexer.token_list[32].type == TokenType.TRUE
    r = "".join([token.text for token in lexer.token_list])
    assert r == string.getvalue()


def test_rule_lexer1(rules):
    string = StringIO(rules[1])
    lexer = Lexer()
    lexer.tokenize(string)

    assert lexer.token_list[0].text == 'isValidDate'
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[1].text == '('
    assert lexer.token_list[1].type == TokenType.Lp
    assert lexer.token_list[2].text == '$*.*.*'
    assert lexer.token_list[2].type == TokenType.Variable
    assert lexer.token_list[3].text == ')'
    assert lexer.token_list[3].type == TokenType.Rp

    r = "".join([token.text for token in lexer.token_list])
    assert r == string.getvalue()


def test_rule_lexer2(rules):
    string = StringIO(rules[2])
    lexer = Lexer()
    lexer.tokenize(string)

    assert lexer.token_list[0].text == 'autoValue'
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[1].text == '('
    assert lexer.token_list[1].type == TokenType.Lp
    assert lexer.token_list[2].text == '('
    assert lexer.token_list[2].type == TokenType.Lp
    assert lexer.token_list[3].text == 'toNum'
    assert lexer.token_list[3].type == TokenType.Identifier
    assert lexer.token_list[4].text == '('
    assert lexer.token_list[4].type == TokenType.Lp
    assert lexer.token_list[5].text == '$*.*.EGORRES.881'
    assert lexer.token_list[5].type == TokenType.Variable
    assert lexer.token_list[6].text == ')'
    assert lexer.token_list[6].type == TokenType.Rp
    assert lexer.token_list[7].text == '!='
    assert lexer.token_list[7].type == TokenType.NEq
    assert lexer.token_list[8].text == '\'\''
    assert lexer.token_list[8].type == TokenType.Identifier
    assert lexer.token_list[9].text == '&&'
    assert lexer.token_list[9].type == TokenType.And
    assert lexer.token_list[10].text == 'toNum'
    assert lexer.token_list[10].type == TokenType.Identifier
    assert lexer.token_list[11].text == '('
    assert lexer.token_list[11].type == TokenType.Lp
    assert lexer.token_list[12].text == '$*.*.EGORRES.879'
    assert lexer.token_list[12].type == TokenType.Variable
    assert lexer.token_list[13].text == ')'
    assert lexer.token_list[13].type == TokenType.Rp
    assert lexer.token_list[14].text == '!='
    assert lexer.token_list[14].type == TokenType.NEq
    assert lexer.token_list[15].text == '\'\''
    assert lexer.token_list[15].type == TokenType.Identifier
    assert lexer.token_list[16].text == '?'
    assert lexer.token_list[16].type == TokenType.Question
    assert lexer.token_list[17].text == 'RoundN'
    assert lexer.token_list[17].type == TokenType.Identifier
    assert lexer.token_list[18].text == '('
    assert lexer.token_list[18].type == TokenType.Lp
    assert lexer.token_list[19].text == 'toNum'
    assert lexer.token_list[19].type == TokenType.Identifier
    assert lexer.token_list[20].text == '('
    assert lexer.token_list[20].type == TokenType.Lp
    assert lexer.token_list[21].text == '$*.*.EGORRES.879'
    assert lexer.token_list[21].type == TokenType.Variable
    assert lexer.token_list[22].text == ')'
    assert lexer.token_list[22].type == TokenType.Rp
    assert lexer.token_list[23].text == '/'
    assert lexer.token_list[23].type == TokenType.Divide
    assert lexer.token_list[24].text == 'Math.pow'
    assert lexer.token_list[24].type == TokenType.Identifier
    assert lexer.token_list[25].text == '('
    assert lexer.token_list[25].type == TokenType.Lp
    assert lexer.token_list[26].text == '('
    assert lexer.token_list[26].type == TokenType.Lp
    assert lexer.token_list[27].text == 'toNum'
    assert lexer.token_list[27].type == TokenType.Identifier
    assert lexer.token_list[28].text == '('
    assert lexer.token_list[28].type == TokenType.Lp
    assert lexer.token_list[29].text == '$*.*.EGORRES.881'
    assert lexer.token_list[29].type == TokenType.Variable
    assert lexer.token_list[30].text == ')'
    assert lexer.token_list[30].type == TokenType.Rp
    assert lexer.token_list[31].text == '/'
    assert lexer.token_list[31].type == TokenType.Divide
    assert lexer.token_list[32].text == '1000'
    assert lexer.token_list[32].type == TokenType.RealNumber
    assert lexer.token_list[33].text == ')'
    assert lexer.token_list[33].type == TokenType.Rp
    assert lexer.token_list[34].text == ','
    assert lexer.token_list[34].type == TokenType.Comma
    assert lexer.token_list[35].text == ' '
    assert lexer.token_list[35].type == TokenType.Blank
    assert lexer.token_list[36].text == '0.33'
    assert lexer.token_list[36].type == TokenType.RealNumber
    assert lexer.token_list[37].text == ')'
    assert lexer.token_list[37].type == TokenType.Rp
    assert lexer.token_list[38].text == ','
    assert lexer.token_list[38].type == TokenType.Comma
    assert lexer.token_list[39].text == ' '
    assert lexer.token_list[39].type == TokenType.Blank
    assert lexer.token_list[40].text == '"1"'
    assert lexer.token_list[40].type == TokenType.Identifier
    assert lexer.token_list[41].text == ')'
    assert lexer.token_list[41].type == TokenType.Rp
    assert lexer.token_list[42].text == ':'
    assert lexer.token_list[42].type == TokenType.Colon
    assert lexer.token_list[43].text == '""'
    assert lexer.token_list[43].type == TokenType.Identifier
    assert lexer.token_list[44].text == ')'
    assert lexer.token_list[44].type == TokenType.Rp
    assert lexer.token_list[45].text == ','
    assert lexer.token_list[45].type == TokenType.Comma
    assert lexer.token_list[46].text == ' '
    assert lexer.token_list[46].type == TokenType.Blank
    assert lexer.token_list[47].text == 'true'
    assert lexer.token_list[47].type == TokenType.TRUE
    assert lexer.token_list[48].text == ')'
    assert lexer.token_list[48].type == TokenType.Rp

    r = "".join([token.text for token in lexer.token_list])
    assert r == string.getvalue()


def test_rule_lexer3(rules):
    string = StringIO(rules[3])
    lexer = Lexer()
    lexer.tokenize(string)

    assert lexer.token_list[0].text == '$*.*.EGPERF'
    assert lexer.token_list[0].type == TokenType.Variable
    assert lexer.token_list[1].text == '=='
    assert lexer.token_list[1].type == TokenType.Eq
    assert lexer.token_list[2].text == '"Y"'
    assert lexer.token_list[2].type == TokenType.Identifier


    r = "".join([token.text for token in lexer.token_list])
    assert r == string.getvalue()


def test_rule_lexer4(rules):
    string = StringIO(rules[4])
    lexer = Lexer()
    lexer.tokenize(string)

    assert lexer.token_list[0].text == '$*.*.EGORRES.12394'
    assert lexer.token_list[0].type == TokenType.Variable
    assert lexer.token_list[1].text == '=='
    assert lexer.token_list[1].type == TokenType.Eq
    assert lexer.token_list[2].text == '"CS"'
    assert lexer.token_list[2].type == TokenType.Identifier
    r = "".join([token.text for token in lexer.token_list])
    assert r == string.getvalue()


def test_rule_lexer5(rules):
    string = StringIO(rules[5])
    lexer = Lexer()
    lexer.tokenize(string)
    assert lexer.token_list[0].text == 'condition'
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[1].text == '('
    assert lexer.token_list[1].type == TokenType.Lp
    assert lexer.token_list[2].text == 'mustAnswer'
    assert lexer.token_list[2].type == TokenType.Identifier
    assert lexer.token_list[3].text == '('
    assert lexer.token_list[3].type == TokenType.Lp
    assert lexer.token_list[4].text == 'toNum'
    assert lexer.token_list[4].type == TokenType.Identifier
    assert lexer.token_list[5].text == '('
    assert lexer.token_list[5].type == TokenType.Lp
    assert lexer.token_list[6].text == '$*.*.*'
    assert lexer.token_list[6].type == TokenType.Variable
    assert lexer.token_list[7].text == ')'
    assert lexer.token_list[7].type == TokenType.Rp
    assert lexer.token_list[8].text == ')'
    assert lexer.token_list[8].type == TokenType.Rp
    assert lexer.token_list[9].text == ','
    assert lexer.token_list[9].type == TokenType.Comma
    assert lexer.token_list[10].text == ' '
    assert lexer.token_list[10].type == TokenType.Blank
    assert lexer.token_list[11].text == 'isRange'
    assert lexer.token_list[11].type == TokenType.Identifier
    assert lexer.token_list[12].text == '('
    assert lexer.token_list[12].type == TokenType.Lp
    assert lexer.token_list[13].text == '$*.*.*'
    assert lexer.token_list[13].type == TokenType.Variable
    assert lexer.token_list[14].text == ','
    assert lexer.token_list[14].type == TokenType.Comma
    assert lexer.token_list[15].text == ' '
    assert lexer.token_list[15].type == TokenType.Blank
    assert lexer.token_list[16].text == '40'
    assert lexer.token_list[16].type == TokenType.RealNumber
    assert lexer.token_list[17].text == ','
    assert lexer.token_list[17].type == TokenType.Comma
    assert lexer.token_list[18].text == '160'
    assert lexer.token_list[18].type == TokenType.RealNumber
    assert lexer.token_list[19].text == ')'
    assert lexer.token_list[19].type == TokenType.Rp
    assert lexer.token_list[20].text == ')'
    assert lexer.token_list[20].type == TokenType.Rp

    r = "".join([token.text for token in lexer.token_list])
    assert r == string.getvalue()


def test_rule_lexer6(rules):
    string = StringIO(rules[6])
    lexer = Lexer()
    lexer.tokenize(string)
    assert lexer.token_list[0].text == 'mustAnswer'
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[1].text == '('
    assert lexer.token_list[1].type == TokenType.Lp
    assert lexer.token_list[2].text == '$*.*.CRONGO'
    assert lexer.token_list[2].type == TokenType.Variable
    assert lexer.token_list[3].text == ')'
    assert lexer.token_list[3].type == TokenType.Rp
    assert lexer.token_list[4].text == '&&'
    assert lexer.token_list[4].type == TokenType.And
    assert lexer.token_list[5].text == '$*.*.CRONGO'
    assert lexer.token_list[5].type == TokenType.Variable
    assert lexer.token_list[6].text == '=='
    assert lexer.token_list[6].type == TokenType.Eq
    assert lexer.token_list[7].text == '\'Y\''
    assert lexer.token_list[7].type == TokenType.Identifier
    assert lexer.token_list[8].text == '?'
    assert lexer.token_list[8].type == TokenType.Question
    assert lexer.token_list[9].text == '$*.*.*'
    assert lexer.token_list[9].type == TokenType.Variable
    assert lexer.token_list[10].text == '=='
    assert lexer.token_list[10].type == TokenType.Eq
    assert lexer.token_list[11].text == '\'\''
    assert lexer.token_list[11].type == TokenType.Identifier
    assert lexer.token_list[12].text == ':'
    assert lexer.token_list[12].type == TokenType.Colon
    assert lexer.token_list[13].text == 'true'
    assert lexer.token_list[13].type == TokenType.TRUE


    r = "".join([token.text for token in lexer.token_list])
    assert r == string.getvalue()


def test_rule_lexer7(rules):
    string = StringIO(rules[7])
    lexer = Lexer()
    lexer.tokenize(string)

    assert lexer.token_list[0].text == 'mustAnswer'
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[1].text == '('
    assert lexer.token_list[1].type == TokenType.Lp
    assert lexer.token_list[2].text == '$*.*.CRONGO'
    assert lexer.token_list[2].type == TokenType.Variable
    assert lexer.token_list[3].text == ')'
    assert lexer.token_list[3].type == TokenType.Rp
    assert lexer.token_list[4].text == '&&'
    assert lexer.token_list[4].type == TokenType.And
    assert lexer.token_list[5].text == '$*.*.CRONGO'
    assert lexer.token_list[5].type == TokenType.Variable
    assert lexer.token_list[6].text == '=='
    assert lexer.token_list[6].type == TokenType.Eq
    assert lexer.token_list[7].text == '\'N\''
    assert lexer.token_list[7].type == TokenType.Identifier
    assert lexer.token_list[8].text == '?'
    assert lexer.token_list[8].type == TokenType.Question
    assert lexer.token_list[9].text == '$*.*.*'
    assert lexer.token_list[9].type == TokenType.Variable
    assert lexer.token_list[10].text == '!='
    assert lexer.token_list[10].type == TokenType.NEq
    assert lexer.token_list[11].text == '\'\''
    assert lexer.token_list[11].type == TokenType.Identifier
    assert lexer.token_list[12].text == ':'
    assert lexer.token_list[12].type == TokenType.Colon
    assert lexer.token_list[13].text == 'true'
    assert lexer.token_list[13].type == TokenType.TRUE
    r = "".join([token.text for token in lexer.token_list])
    assert r == string.getvalue()


def test_rule_lexer8(rules):
    string = StringIO(rules[8])
    lexer = Lexer()
    lexer.tokenize(string)

    assert len(lexer.token_list) == 10
    r = "".join([token.text for token in lexer.token_list])
    assert r == string.getvalue()


def test_rule_lexer9(rules):
    string = StringIO(rules[4])
    lexer = Lexer()
    lexer.tokenize(string)

    assert len(lexer.token_list) == 3
    r = "".join([token.text for token in lexer.token_list])
    assert r == string.getvalue()


def test_rule_lexer10(rules):
    string = StringIO(rules[10])
    lexer = Lexer()
    lexer.tokenize(string)
    # dipsplay(string, lexer)
    assert len(lexer.token_list) == 48
    r = "".join([token.text for token in lexer.token_list])
    assert r == string.getvalue()


def test_rule_lexer11(rules):
    string = StringIO(rules[11])
    lexer = Lexer()
    lexer.tokenize(string)

    assert lexer.token_list[0].text == 'toDate'
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[1].text == '('
    assert lexer.token_list[1].type == TokenType.Lp
    assert lexer.token_list[2].text == '$C1D1.ONC-392 Administration.ECSTDAT'
    assert lexer.token_list[2].type == TokenType.Variable
    assert lexer.token_list[3].text == ')'
    assert lexer.token_list[3].type == TokenType.Rp
    assert lexer.token_list[4].text == '!='
    assert lexer.token_list[4].type == TokenType.NEq
    assert lexer.token_list[5].text == '""'
    assert lexer.token_list[5].type == TokenType.Identifier
    assert lexer.token_list[6].text == '&&'
    assert lexer.token_list[6].type == TokenType.And
    assert lexer.token_list[7].text == ' '
    assert lexer.token_list[7].type == TokenType.Blank
    assert lexer.token_list[8].text == 'toDate'
    assert lexer.token_list[8].type == TokenType.Identifier
    assert lexer.token_list[9].text == '('
    assert lexer.token_list[9].type == TokenType.Lp
    assert lexer.token_list[10].text == '$*.*.LBDAT'
    assert lexer.token_list[10].type == TokenType.Variable
    assert lexer.token_list[11].text == ')'
    assert lexer.token_list[11].type == TokenType.Rp
    assert lexer.token_list[12].text == '!='
    assert lexer.token_list[12].type == TokenType.NEq
    assert lexer.token_list[13].text == '""'
    assert lexer.token_list[13].type == TokenType.Identifier
    assert lexer.token_list[14].text == '&&'
    assert lexer.token_list[14].type == TokenType.And
    assert lexer.token_list[15].text == ' '
    assert lexer.token_list[15].type == TokenType.Blank
    assert lexer.token_list[16].text == 'toDate'
    assert lexer.token_list[16].type == TokenType.Identifier
    assert lexer.token_list[17].text == '('
    assert lexer.token_list[17].type == TokenType.Lp
    assert lexer.token_list[18].text == '"2015-1-12 "'
    assert lexer.token_list[18].type == TokenType.Identifier
    assert lexer.token_list[19].text == '+'
    assert lexer.token_list[19].type == TokenType.Plus
    assert lexer.token_list[20].text == '$C1D1.ONC-392 Administration.ECSTTIM'
    assert lexer.token_list[20].type == TokenType.Variable
    assert lexer.token_list[21].text == '+'
    assert lexer.token_list[21].type == TokenType.Plus
    assert lexer.token_list[22].text == '":00"'
    assert lexer.token_list[22].type == TokenType.Identifier
    assert lexer.token_list[23].text == ')'
    assert lexer.token_list[23].type == TokenType.Rp
    assert lexer.token_list[24].text == '!='
    assert lexer.token_list[24].type == TokenType.NEq
    assert lexer.token_list[25].text == '""'
    assert lexer.token_list[25].type == TokenType.Identifier
    assert lexer.token_list[26].text == ' '
    assert lexer.token_list[26].type == TokenType.Blank
    assert lexer.token_list[27].text == '&&'
    assert lexer.token_list[27].type == TokenType.And
    assert lexer.token_list[28].text == 'toDate'
    assert lexer.token_list[28].type == TokenType.Identifier
    assert lexer.token_list[29].text == '('
    assert lexer.token_list[29].type == TokenType.Lp
    assert lexer.token_list[30].text == '"2015-1-12 "'
    assert lexer.token_list[30].type == TokenType.Identifier
    assert lexer.token_list[31].text == '+'
    assert lexer.token_list[31].type == TokenType.Plus
    assert lexer.token_list[32].text == '$*.*.*'
    assert lexer.token_list[32].type == TokenType.Variable
    assert lexer.token_list[33].text == '+'
    assert lexer.token_list[33].type == TokenType.Plus
    assert lexer.token_list[34].text == '":00"'
    assert lexer.token_list[34].type == TokenType.Identifier
    assert lexer.token_list[35].text == ')'
    assert lexer.token_list[35].type == TokenType.Rp
    assert lexer.token_list[36].text == '!='
    assert lexer.token_list[36].type == TokenType.NEq
    assert lexer.token_list[37].text == '""'
    assert lexer.token_list[37].type == TokenType.Identifier
    assert lexer.token_list[38].text == '?'
    assert lexer.token_list[38].type == TokenType.Question
    assert lexer.token_list[39].text == 'dateDiff'
    assert lexer.token_list[39].type == TokenType.Identifier
    assert lexer.token_list[40].text == '('
    assert lexer.token_list[40].type == TokenType.Lp
    assert lexer.token_list[41].text == '$C1D1.ONC-392 Administration.ECSTDAT'
    assert lexer.token_list[41].type == TokenType.Variable
    assert lexer.token_list[42].text == '+'
    assert lexer.token_list[42].type == TokenType.Plus
    assert lexer.token_list[43].text == '" "'
    assert lexer.token_list[43].type == TokenType.Identifier
    assert lexer.token_list[44].text == '+'
    assert lexer.token_list[44].type == TokenType.Plus
    assert lexer.token_list[45].text == '$C1D1.ONC-392 Administration.ECSTTIM'
    assert lexer.token_list[45].type == TokenType.Variable
    assert lexer.token_list[46].text == '+'
    assert lexer.token_list[46].type == TokenType.Plus
    assert lexer.token_list[47].text == '":00"'
    assert lexer.token_list[47].type == TokenType.Identifier
    assert lexer.token_list[48].text == ','
    assert lexer.token_list[48].type == TokenType.Comma
    assert lexer.token_list[49].text == '$*.*.LBDAT'
    assert lexer.token_list[49].type == TokenType.Variable
    assert lexer.token_list[50].text == '+'
    assert lexer.token_list[50].type == TokenType.Plus
    assert lexer.token_list[51].text == '" "'
    assert lexer.token_list[51].type == TokenType.Identifier
    assert lexer.token_list[52].text == '+'
    assert lexer.token_list[52].type == TokenType.Plus
    assert lexer.token_list[53].text == '$*.*.*'
    assert lexer.token_list[53].type == TokenType.Variable
    assert lexer.token_list[54].text == '+'
    assert lexer.token_list[54].type == TokenType.Plus
    assert lexer.token_list[55].text == '":00"'
    assert lexer.token_list[55].type == TokenType.Identifier
    assert lexer.token_list[56].text == ','
    assert lexer.token_list[56].type == TokenType.Comma
    assert lexer.token_list[57].text == '"m"'
    assert lexer.token_list[57].type == TokenType.Identifier
    assert lexer.token_list[58].text == ')'
    assert lexer.token_list[58].type == TokenType.Rp
    assert lexer.token_list[59].text == '>'
    assert lexer.token_list[59].type == TokenType.Gt
    assert lexer.token_list[60].text == '0'
    assert lexer.token_list[60].type == TokenType.RealNumber
    assert lexer.token_list[61].text == ':'
    assert lexer.token_list[61].type == TokenType.Colon
    assert lexer.token_list[62].text == 'true'
    assert lexer.token_list[62].type == TokenType.TRUE
    r = "".join([token.text for token in lexer.token_list])
    assert r == string.getvalue()


def test_rule_lexer12(rules):
    string = StringIO(rules[12])
    lexer = Lexer()
    lexer.tokenize(string)

    assert lexer.token_list[0].text == 'autoValue'
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[1].text == '('
    assert lexer.token_list[1].type == TokenType.Lp
    assert lexer.token_list[2].text == 'RoundN'
    assert lexer.token_list[2].type == TokenType.Identifier
    assert lexer.token_list[3].text == '('
    assert lexer.token_list[3].type == TokenType.Lp
    assert lexer.token_list[4].text == 'sum'
    assert lexer.token_list[4].type == TokenType.Identifier
    assert lexer.token_list[5].text == '('
    assert lexer.token_list[5].type == TokenType.Lp
    assert lexer.token_list[6].text == 'getSumOfItemInLog'
    assert lexer.token_list[6].type == TokenType.Identifier
    assert lexer.token_list[7].text == '('
    assert lexer.token_list[7].type == TokenType.Lp
    assert lexer.token_list[8].text == '$*.Target Lesions Assessment (Details) (Screening).TLLONG'
    assert lexer.token_list[8].type == TokenType.Variable
    assert lexer.token_list[9].text == ','
    assert lexer.token_list[9].type == TokenType.Comma
    assert lexer.token_list[10].text == ' '
    assert lexer.token_list[10].type == TokenType.Blank
    assert lexer.token_list[11].text == '$*.Target Lesions Assessment (Details) (Screening).TLLOC '
    assert lexer.token_list[11].type == TokenType.Variable
    assert lexer.token_list[12].text == '!='
    assert lexer.token_list[12].type == TokenType.NEq
    assert lexer.token_list[13].text == ' '
    assert lexer.token_list[13].type == TokenType.Blank
    assert lexer.token_list[14].text == '11'
    assert lexer.token_list[14].type == TokenType.RealNumber
    assert lexer.token_list[15].text == ' '
    assert lexer.token_list[15].type == TokenType.Blank
    assert lexer.token_list[16].text == '&&'
    assert lexer.token_list[16].type == TokenType.And
    assert lexer.token_list[17].text == ' '
    assert lexer.token_list[17].type == TokenType.Blank
    assert lexer.token_list[18].text == '$*.Target Lesions Assessment (Details) (Screening).TLDIAU '
    assert lexer.token_list[18].type == TokenType.Variable
    assert lexer.token_list[19].text == '=='
    assert lexer.token_list[19].type == TokenType.Eq
    assert lexer.token_list[20].text == ' '
    assert lexer.token_list[20].type == TokenType.Blank
    assert lexer.token_list[21].text == '"CM"'
    assert lexer.token_list[21].type == TokenType.Identifier
    assert lexer.token_list[22].text == ')'
    assert lexer.token_list[22].type == TokenType.Rp
    assert lexer.token_list[23].text == ' '
    assert lexer.token_list[23].type == TokenType.Blank
    assert lexer.token_list[24].text == ','
    assert lexer.token_list[24].type == TokenType.Comma
    assert lexer.token_list[25].text == ' '
    assert lexer.token_list[25].type == TokenType.Blank
    assert lexer.token_list[26].text == 'multiply'
    assert lexer.token_list[26].type == TokenType.Identifier
    assert lexer.token_list[27].text == '('
    assert lexer.token_list[27].type == TokenType.Lp
    assert lexer.token_list[28].text == 'getSumOfItemInLog'
    assert lexer.token_list[28].type == TokenType.Identifier
    assert lexer.token_list[29].text == '('
    assert lexer.token_list[29].type == TokenType.Lp
    assert lexer.token_list[30].text == '$*.Target Lesions Assessment (Details) (Screening).TLLONG'
    assert lexer.token_list[30].type == TokenType.Variable
    assert lexer.token_list[31].text == ','
    assert lexer.token_list[31].type == TokenType.Comma
    assert lexer.token_list[32].text == ' '
    assert lexer.token_list[32].type == TokenType.Blank
    assert lexer.token_list[33].text == '$*.Target Lesions Assessment (Details) (Screening).TLLOC '
    assert lexer.token_list[33].type == TokenType.Variable
    assert lexer.token_list[34].text == '!='
    assert lexer.token_list[34].type == TokenType.NEq
    assert lexer.token_list[35].text == ' '
    assert lexer.token_list[35].type == TokenType.Blank
    assert lexer.token_list[36].text == '11'
    assert lexer.token_list[36].type == TokenType.RealNumber
    assert lexer.token_list[37].text == ' '
    assert lexer.token_list[37].type == TokenType.Blank
    assert lexer.token_list[38].text == '&&'
    assert lexer.token_list[38].type == TokenType.And
    assert lexer.token_list[39].text == ' '
    assert lexer.token_list[39].type == TokenType.Blank
    assert lexer.token_list[40].text == '$*.Target Lesions Assessment (Details) (Screening).TLDIAU '
    assert lexer.token_list[40].type == TokenType.Variable
    assert lexer.token_list[41].text == '=='
    assert lexer.token_list[41].type == TokenType.Eq
    assert lexer.token_list[42].text == ' '
    assert lexer.token_list[42].type == TokenType.Blank
    assert lexer.token_list[43].text == '"MM"'
    assert lexer.token_list[43].type == TokenType.Identifier
    assert lexer.token_list[44].text == ')'
    assert lexer.token_list[44].type == TokenType.Rp
    assert lexer.token_list[45].text == ','
    assert lexer.token_list[45].type == TokenType.Comma
    assert lexer.token_list[46].text == '1'
    assert lexer.token_list[46].type == TokenType.RealNumber
    assert lexer.token_list[47].text == '/'
    assert lexer.token_list[47].type == TokenType.Divide
    assert lexer.token_list[48].text == '10'
    assert lexer.token_list[48].type == TokenType.RealNumber
    assert lexer.token_list[49].text == ')'
    assert lexer.token_list[49].type == TokenType.Rp
    assert lexer.token_list[50].text == ' '
    assert lexer.token_list[50].type == TokenType.Blank
    assert lexer.token_list[51].text == ','
    assert lexer.token_list[51].type == TokenType.Comma
    assert lexer.token_list[52].text == ' '
    assert lexer.token_list[52].type == TokenType.Blank
    assert lexer.token_list[53].text == 'getSumOfItemInLog'
    assert lexer.token_list[53].type == TokenType.Identifier
    assert lexer.token_list[54].text == '('
    assert lexer.token_list[54].type == TokenType.Lp
    assert lexer.token_list[55].text == '$*.Target Lesions Assessment (Details) (Screening).TLSHORT'
    assert lexer.token_list[55].type == TokenType.Variable
    assert lexer.token_list[56].text == ','
    assert lexer.token_list[56].type == TokenType.Comma
    assert lexer.token_list[57].text == ' '
    assert lexer.token_list[57].type == TokenType.Blank
    assert lexer.token_list[58].text == '$*.Target Lesions Assessment (Details) (Screening).TLLOC '
    assert lexer.token_list[58].type == TokenType.Variable
    assert lexer.token_list[59].text == '=='
    assert lexer.token_list[59].type == TokenType.Eq
    assert lexer.token_list[60].text == ' '
    assert lexer.token_list[60].type == TokenType.Blank
    assert lexer.token_list[61].text == '11'
    assert lexer.token_list[61].type == TokenType.RealNumber
    assert lexer.token_list[62].text == ' '
    assert lexer.token_list[62].type == TokenType.Blank
    assert lexer.token_list[63].text == '&&'
    assert lexer.token_list[63].type == TokenType.And
    assert lexer.token_list[64].text == ' '
    assert lexer.token_list[64].type == TokenType.Blank
    assert lexer.token_list[65].text == '$*.Target Lesions Assessment (Details) (Screening).TLDIAU'
    assert lexer.token_list[65].type == TokenType.Variable
    assert lexer.token_list[66].text == '=='
    assert lexer.token_list[66].type == TokenType.Eq
    assert lexer.token_list[67].text == ' '
    assert lexer.token_list[67].type == TokenType.Blank
    assert lexer.token_list[68].text == '"CM"'
    assert lexer.token_list[68].type == TokenType.Identifier
    assert lexer.token_list[69].text == ')'
    assert lexer.token_list[69].type == TokenType.Rp
    assert lexer.token_list[70].text == ' '
    assert lexer.token_list[70].type == TokenType.Blank
    assert lexer.token_list[71].text == ','
    assert lexer.token_list[71].type == TokenType.Comma
    assert lexer.token_list[72].text == ' '
    assert lexer.token_list[72].type == TokenType.Blank
    assert lexer.token_list[73].text == 'multiply'
    assert lexer.token_list[73].type == TokenType.Identifier
    assert lexer.token_list[74].text == '('
    assert lexer.token_list[74].type == TokenType.Lp
    assert lexer.token_list[75].text == 'getSumOfItemInLog'
    assert lexer.token_list[75].type == TokenType.Identifier
    assert lexer.token_list[76].text == '('
    assert lexer.token_list[76].type == TokenType.Lp
    assert lexer.token_list[77].text == '$*.Target Lesions Assessment (Details) (Screening).TLSHORT'
    assert lexer.token_list[77].type == TokenType.Variable
    assert lexer.token_list[78].text == ','
    assert lexer.token_list[78].type == TokenType.Comma
    assert lexer.token_list[79].text == ' '
    assert lexer.token_list[79].type == TokenType.Blank
    assert lexer.token_list[80].text == '$*.Target Lesions Assessment (Details) (Screening).TLLOC '
    assert lexer.token_list[80].type == TokenType.Variable
    assert lexer.token_list[81].text == '=='
    assert lexer.token_list[81].type == TokenType.Eq
    assert lexer.token_list[82].text == ' '
    assert lexer.token_list[82].type == TokenType.Blank
    assert lexer.token_list[83].text == '11'
    assert lexer.token_list[83].type == TokenType.RealNumber
    assert lexer.token_list[84].text == ' '
    assert lexer.token_list[84].type == TokenType.Blank
    assert lexer.token_list[85].text == '&&'
    assert lexer.token_list[85].type == TokenType.And
    assert lexer.token_list[86].text == ' '
    assert lexer.token_list[86].type == TokenType.Blank
    assert lexer.token_list[87].text == '$*.Target Lesions Assessment (Details) (Screening).TLDIAU'
    assert lexer.token_list[87].type == TokenType.Variable
    assert lexer.token_list[88].text == '=='
    assert lexer.token_list[88].type == TokenType.Eq
    assert lexer.token_list[89].text == ' '
    assert lexer.token_list[89].type == TokenType.Blank
    assert lexer.token_list[90].text == '"MM"'
    assert lexer.token_list[90].type == TokenType.Identifier
    assert lexer.token_list[91].text == ')'
    assert lexer.token_list[91].type == TokenType.Rp
    assert lexer.token_list[92].text == ','
    assert lexer.token_list[92].type == TokenType.Comma
    assert lexer.token_list[93].text == '1'
    assert lexer.token_list[93].type == TokenType.RealNumber
    assert lexer.token_list[94].text == '/'
    assert lexer.token_list[94].type == TokenType.Divide
    assert lexer.token_list[95].text == '10'
    assert lexer.token_list[95].type == TokenType.RealNumber
    assert lexer.token_list[96].text == ')'
    assert lexer.token_list[96].type == TokenType.Rp
    assert lexer.token_list[97].text == ')'
    assert lexer.token_list[97].type == TokenType.Rp
    assert lexer.token_list[98].text == ','
    assert lexer.token_list[98].type == TokenType.Comma
    assert lexer.token_list[99].text == ' '
    assert lexer.token_list[99].type == TokenType.Blank
    assert lexer.token_list[100].text == '0.01'
    assert lexer.token_list[100].type == TokenType.RealNumber
    assert lexer.token_list[101].text == ')'
    assert lexer.token_list[101].type == TokenType.Rp
    assert lexer.token_list[102].text == ','
    assert lexer.token_list[102].type == TokenType.Comma
    assert lexer.token_list[103].text == ' '
    assert lexer.token_list[103].type == TokenType.Blank
    assert lexer.token_list[104].text == 'true'
    assert lexer.token_list[104].type == TokenType.TRUE
    assert lexer.token_list[105].text == ')'
    assert lexer.token_list[105].type == TokenType.Rp
    r = "".join([token.text for token in lexer.token_list])
    assert r == string.getvalue()


def display(lexer):
    for i, t in enumerate(lexer.token_list):
        print("assert", f'lexer.token_list[{i}].text == \'{t.text}\'')
        print("assert", f'lexer.token_list[{i}].type == {t.type}')



from io import StringIO

import pytest

from edkrule.interpreter.lexer.lexer import Lexer
from edkrule.interpreter.lexer.token_type import TokenType


def test_lexer():
    lexer = Lexer()
    identifier_token(lexer)
    true_token(lexer)
    blank_token_with_ture_false(lexer)
    blank_token_with_id(lexer)
    comma_with_true_false(lexer)
    byte_and_token(lexer)
    and_token(lexer)
    byte_or_token(lexer)
    or_token(lexer)
    real_number_token(lexer)
    plus_token(lexer)
    minus_token(lexer)
    d_plus_token(lexer)
    d_minus_token(lexer)
    complex_expression(lexer)
    plus_eq_token(lexer)
    minus_eq_token(lexer)
    colon_token(lexer)
    sq_token(lexer)
    dq_token(lexer)
    comma_token(lexer)
    question_token(lexer)
    lp_rp_token(lexer)
    test_vars()
    test_ge_and_gt()
    neq()
    le_and_lt()
    assign()
    eq_plus()
    eq_minus()
    aways_eq_plus()
    mul_and_divide()
    test_math_name()



def identifier_token(lexer: Lexer):
    string = StringIO("t")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[0].text == "t"
    string = StringIO("tr")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[0].text == "tr"
    string = StringIO("tru")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[0].text == "tru"
    string = StringIO("trua")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[0].text == "trua"
    string = StringIO("f")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[0].text == "f"
    string = StringIO("fa")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[0].text == "fa"
    string = StringIO("fal")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[0].text == "fal"
    string = StringIO("fals")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[0].text == "fals"
    string = StringIO("falsg")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[0].text == "falsg"
    string = StringIO("aaaa1111")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[0].text == "aaaa1111"


def true_token(lexer: Lexer):
    string = StringIO("true")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.TRUE
    assert lexer.token_list[0].text == "true"
    string = StringIO("abcdtrue")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[0].text == "abcdtrue"
    string = StringIO("abcdtrus")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[0].text == "abcdtrus"
    string = StringIO("true111")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[0].text == "true111"


def false_token(lexer: Lexer):
    string = StringIO("false")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.FALSE
    assert lexer.token_list[0].text == "false"
    string = StringIO("falseaasds")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[0].text == "falseaasds"


def blank_token_with_id(lexer: Lexer):
    string = StringIO("a b")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[0].text == "a"
    assert lexer.token_list[1].type == TokenType.Blank
    assert lexer.token_list[1].text == " "
    assert lexer.token_list[2].type == TokenType.Identifier
    assert lexer.token_list[2].text == "b"


def blank_token_with_ture_false(lexer: Lexer):
    string = StringIO("true false")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.TRUE
    assert lexer.token_list[0].text == "true"
    assert lexer.token_list[1].type == TokenType.Blank
    assert lexer.token_list[1].text == " "
    assert lexer.token_list[2].type == TokenType.FALSE
    assert lexer.token_list[2].text == "false"
    string = StringIO("true xxxxxxxxxx false false ")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.TRUE
    assert lexer.token_list[0].text == "true"
    assert lexer.token_list[1].type == TokenType.Blank
    assert lexer.token_list[1].text == " "
    assert lexer.token_list[2].type == TokenType.Identifier
    assert lexer.token_list[2].text == "xxxxxxxxxx"
    assert lexer.token_list[3].type == TokenType.Blank
    assert lexer.token_list[3].text == " "
    assert lexer.token_list[4].type == TokenType.FALSE
    assert lexer.token_list[4].text == "false"
    assert lexer.token_list[5].type == TokenType.Blank
    assert lexer.token_list[5].text == " "
    assert lexer.token_list[6].type == TokenType.FALSE
    assert lexer.token_list[6].text == "false"
    string = StringIO("true xxxxxxxxxx false true ")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.TRUE
    assert lexer.token_list[0].text == "true"
    assert lexer.token_list[1].type == TokenType.Blank
    assert lexer.token_list[1].text == " "
    assert lexer.token_list[2].type == TokenType.Identifier
    assert lexer.token_list[2].text == "xxxxxxxxxx"
    assert lexer.token_list[3].type == TokenType.Blank
    assert lexer.token_list[3].text == " "
    assert lexer.token_list[4].type == TokenType.FALSE
    assert lexer.token_list[4].text == "false"
    assert lexer.token_list[5].type == TokenType.Blank
    assert lexer.token_list[5].text == " "
    assert lexer.token_list[6].type == TokenType.TRUE
    assert lexer.token_list[6].text == "true"


def byte_and_token(lexer: Lexer):
    string = StringIO("&")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.ByteAnd
    assert lexer.token_list[0].text == "&"
    string = StringIO("true & false")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.TRUE
    assert lexer.token_list[0].text == "true"
    assert lexer.token_list[1].type == TokenType.Blank
    assert lexer.token_list[1].text == " "
    assert lexer.token_list[2].type == TokenType.ByteAnd
    assert lexer.token_list[2].text == "&"
    assert lexer.token_list[3].type == TokenType.Blank
    assert lexer.token_list[3].text == " "
    assert lexer.token_list[4].type == TokenType.FALSE
    assert lexer.token_list[4].text == "false"
    string = StringIO("true&false")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.TRUE
    assert lexer.token_list[0].text == "true"
    assert lexer.token_list[1].type == TokenType.ByteAnd
    assert lexer.token_list[1].text == "&"
    assert lexer.token_list[2].type == TokenType.FALSE
    assert lexer.token_list[2].text == "false"
    string = StringIO("true&fgggg")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.TRUE
    assert lexer.token_list[0].text == "true"
    assert lexer.token_list[1].type == TokenType.ByteAnd
    assert lexer.token_list[1].text == "&"
    assert lexer.token_list[2].type == TokenType.Identifier
    assert lexer.token_list[2].text == "fgggg"


def comma_with_true_false(lexer: Lexer):
    string = StringIO("true,false")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.TRUE
    assert lexer.token_list[1].type == TokenType.Comma
    assert lexer.token_list[2].type == TokenType.FALSE

def and_token(lexer: Lexer):
    string = StringIO("&&")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.And
    assert lexer.token_list[0].text == "&&"
    string = StringIO("true&&false")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.TRUE
    assert lexer.token_list[0].text == "true"
    assert lexer.token_list[1].type == TokenType.And
    assert lexer.token_list[1].text == "&&"
    assert lexer.token_list[2].type == TokenType.FALSE
    assert lexer.token_list[2].text == "false"
    string = StringIO("true && false")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.TRUE
    assert lexer.token_list[0].text == "true"
    assert lexer.token_list[1].type == TokenType.Blank
    assert lexer.token_list[1].text == " "
    assert lexer.token_list[2].type == TokenType.And
    assert lexer.token_list[2].text == "&&"
    assert lexer.token_list[3].type == TokenType.Blank
    assert lexer.token_list[3].text == " "
    assert lexer.token_list[4].type == TokenType.FALSE
    assert lexer.token_list[4].text == "false"
    string = StringIO("true&&fgggg")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.TRUE
    assert lexer.token_list[0].text == "true"
    assert lexer.token_list[1].type == TokenType.And
    assert lexer.token_list[1].text == "&&"
    assert lexer.token_list[2].type == TokenType.Identifier
    assert lexer.token_list[2].text == "fgggg"


def byte_or_token(lexer: Lexer):
    string = StringIO("|")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.ByteOr
    assert lexer.token_list[0].text == "|"
    string = StringIO("true | false")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.TRUE
    assert lexer.token_list[0].text == "true"
    assert lexer.token_list[1].type == TokenType.Blank
    assert lexer.token_list[1].text == " "
    assert lexer.token_list[2].type == TokenType.ByteOr
    assert lexer.token_list[2].text == "|"
    assert lexer.token_list[3].type == TokenType.Blank
    assert lexer.token_list[3].text == " "
    assert lexer.token_list[4].type == TokenType.FALSE
    assert lexer.token_list[4].text == "false"
    string = StringIO("true|false")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.TRUE
    assert lexer.token_list[0].text == "true"
    assert lexer.token_list[1].type == TokenType.ByteOr
    assert lexer.token_list[1].text == "|"
    assert lexer.token_list[2].type == TokenType.FALSE
    assert lexer.token_list[2].text == "false"
    string = StringIO("true|fgggg")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.TRUE
    assert lexer.token_list[0].text == "true"
    assert lexer.token_list[1].type == TokenType.ByteOr
    assert lexer.token_list[1].text == "|"
    assert lexer.token_list[2].type == TokenType.Identifier
    assert lexer.token_list[2].text == "fgggg"


def or_token(lexer: Lexer):
    string = StringIO("||")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Or
    assert lexer.token_list[0].text == "||"
    string = StringIO("true || false")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.TRUE
    assert lexer.token_list[0].text == "true"
    assert lexer.token_list[1].type == TokenType.Blank
    assert lexer.token_list[1].text == " "
    assert lexer.token_list[2].type == TokenType.Or
    assert lexer.token_list[2].text == "||"
    assert lexer.token_list[3].type == TokenType.Blank
    assert lexer.token_list[3].text == " "
    assert lexer.token_list[4].type == TokenType.FALSE
    assert lexer.token_list[4].text == "false"
    string = StringIO("true||false")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.TRUE
    assert lexer.token_list[0].text == "true"
    assert lexer.token_list[1].type == TokenType.Or
    assert lexer.token_list[1].text == "||"
    assert lexer.token_list[2].type == TokenType.FALSE
    assert lexer.token_list[2].text == "false"
    string = StringIO("true||fgggg")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.TRUE
    assert lexer.token_list[0].text == "true"
    assert lexer.token_list[1].type == TokenType.Or
    assert lexer.token_list[1].text == "||"
    assert lexer.token_list[2].type == TokenType.Identifier
    assert lexer.token_list[2].text == "fgggg"


def real_number_token(lexer: Lexer):
    string = StringIO("1")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.RealNumber
    assert lexer.token_list[0].text == "1"
    string = StringIO("1.2")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.RealNumber
    assert lexer.token_list[0].text == "1.2"
    string = StringIO("xxx 1.2 true")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[0].text == "xxx"
    assert lexer.token_list[1].type == TokenType.Blank
    assert lexer.token_list[1].text == " "
    assert lexer.token_list[2].type == TokenType.RealNumber
    assert lexer.token_list[2].text == "1.2"
    assert lexer.token_list[3].type == TokenType.Blank
    assert lexer.token_list[3].text == " "
    assert lexer.token_list[4].type == TokenType.TRUE
    assert lexer.token_list[4].text == "true"
    string = StringIO("0.10")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.RealNumber
    assert lexer.token_list[0].text == "0.10"
    with pytest.raises(Exception):
        string = StringIO("xxx 1. true")
        lexer.tokenize(string)


def plus_token(lexer: Lexer):
    string = StringIO("1+2")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.RealNumber
    assert lexer.token_list[0].text == "1"
    assert lexer.token_list[1].type == TokenType.Plus
    assert lexer.token_list[1].text == "+"
    assert lexer.token_list[2].type == TokenType.RealNumber
    assert lexer.token_list[2].text == "2"
    string = StringIO("1+aaaaa")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.RealNumber
    assert lexer.token_list[0].text == "1"
    assert lexer.token_list[1].type == TokenType.Plus
    assert lexer.token_list[1].text == "+"
    assert lexer.token_list[2].type == TokenType.Identifier
    assert lexer.token_list[2].text == "aaaaa"
    string = StringIO("+")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Plus
    assert lexer.token_list[0].text == "+"


def minus_token(lexer: Lexer):
    string = StringIO("1-2")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.RealNumber
    assert lexer.token_list[0].text == "1"
    assert lexer.token_list[1].type == TokenType.Minus
    assert lexer.token_list[1].text == "-"
    assert lexer.token_list[2].type == TokenType.RealNumber
    assert lexer.token_list[2].text == "2"
    string = StringIO("1-aaaaa")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.RealNumber
    assert lexer.token_list[0].text == "1"
    assert lexer.token_list[1].type == TokenType.Minus
    assert lexer.token_list[1].text == "-"
    assert lexer.token_list[2].type == TokenType.Identifier
    assert lexer.token_list[2].text == "aaaaa"
    string = StringIO("-")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Minus
    assert lexer.token_list[0].text == "-"


def d_plus_token(lexer: Lexer):
    string = StringIO("++")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.DPlus
    assert lexer.token_list[0].text == "++"
    string = StringIO("1++")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.RealNumber
    assert lexer.token_list[0].text == "1"
    assert lexer.token_list[1].type == TokenType.DPlus
    assert lexer.token_list[1].text == "++"

    string = StringIO("a++")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[0].text == "a"
    assert lexer.token_list[1].type == TokenType.DPlus
    assert lexer.token_list[1].text == "++"

    string = StringIO("a++b")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[0].text == "a"
    assert lexer.token_list[1].type == TokenType.DPlus
    assert lexer.token_list[1].text == "++"
    assert lexer.token_list[2].type == TokenType.Identifier
    assert lexer.token_list[2].text == "b"


def d_minus_token(lexer: Lexer):
    string = StringIO("--")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.DMinus
    assert lexer.token_list[0].text == "--"
    string = StringIO("1--")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.RealNumber
    assert lexer.token_list[0].text == "1"
    assert lexer.token_list[1].type == TokenType.DMinus
    assert lexer.token_list[1].text == "--"

    string = StringIO("a--")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[0].text == "a"
    assert lexer.token_list[1].type == TokenType.DMinus
    assert lexer.token_list[1].text == "--"

    string = StringIO("a--b")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[0].text == "a"
    assert lexer.token_list[1].type == TokenType.DMinus
    assert lexer.token_list[1].text == "--"
    assert lexer.token_list[2].type == TokenType.Identifier
    assert lexer.token_list[2].text == "b"


def complex_expression(lexer: Lexer):
    string = StringIO("1++1+aa1s-d-- ")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.RealNumber
    assert lexer.token_list[0].text == "1"
    assert lexer.token_list[1].type == TokenType.DPlus
    assert lexer.token_list[1].text == "++"
    assert lexer.token_list[2].type == TokenType.RealNumber
    assert lexer.token_list[2].text == "1"
    assert lexer.token_list[3].type == TokenType.Plus
    assert lexer.token_list[3].text == "+"
    assert lexer.token_list[4].type == TokenType.Identifier
    assert lexer.token_list[4].text == "aa1s"
    assert lexer.token_list[5].type == TokenType.Minus
    assert lexer.token_list[5].text == "-"
    assert lexer.token_list[6].type == TokenType.Identifier
    assert lexer.token_list[6].text == "d"
    assert lexer.token_list[7].type == TokenType.DMinus
    assert lexer.token_list[7].text == "--"


def plus_eq_token(lexer: Lexer):
    string = StringIO("+=")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.PlusEq
    assert lexer.token_list[0].text == "+="
    string = StringIO("a+=1")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[0].text == "a"
    assert lexer.token_list[1].type == TokenType.PlusEq
    assert lexer.token_list[1].text == "+="
    assert lexer.token_list[2].type == TokenType.RealNumber
    assert lexer.token_list[2].text == "1"


def minus_eq_token(lexer: Lexer):
    string = StringIO("-=")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.MinusEq
    assert lexer.token_list[0].text == "-="
    string = StringIO("a-=1")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[0].text == "a"
    assert lexer.token_list[1].type == TokenType.MinusEq
    assert lexer.token_list[1].text == "-="
    assert lexer.token_list[2].type == TokenType.RealNumber
    assert lexer.token_list[2].text == "1"


def colon_token(lexer: Lexer):
    string = StringIO(":")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Colon
    assert lexer.token_list[0].text == ":"
    string = StringIO("1:1")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.RealNumber
    assert lexer.token_list[0].text == "1"
    assert lexer.token_list[1].type == TokenType.Colon
    assert lexer.token_list[1].text == ":"
    assert lexer.token_list[2].type == TokenType.RealNumber
    assert lexer.token_list[2].text == "1"
    string = StringIO(":1+1:1++:ad:true:false:")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Colon
    assert lexer.token_list[0].text == ":"
    assert lexer.token_list[1].type == TokenType.RealNumber
    assert lexer.token_list[1].text == "1"
    assert lexer.token_list[2].type == TokenType.Plus
    assert lexer.token_list[2].text == "+"
    assert lexer.token_list[3].type == TokenType.RealNumber
    assert lexer.token_list[3].text == "1"
    assert lexer.token_list[4].type == TokenType.Colon
    assert lexer.token_list[4].text == ":"
    assert lexer.token_list[5].type == TokenType.RealNumber
    assert lexer.token_list[5].text == "1"
    assert lexer.token_list[6].type == TokenType.DPlus
    assert lexer.token_list[6].text == "++"

    assert lexer.token_list[7].type == TokenType.Colon
    assert lexer.token_list[7].text == ":"
    assert lexer.token_list[8].type == TokenType.Identifier
    assert lexer.token_list[8].text == "ad"
    assert lexer.token_list[9].type == TokenType.Colon
    assert lexer.token_list[9].text == ":"
    assert lexer.token_list[10].type == TokenType.TRUE
    assert lexer.token_list[10].text == "true"
    assert lexer.token_list[11].type == TokenType.Colon
    assert lexer.token_list[11].text == ":"
    assert lexer.token_list[12].type == TokenType.FALSE
    assert lexer.token_list[12].text == "false"
    assert lexer.token_list[13].type == TokenType.Colon
    assert lexer.token_list[13].text == ":"


def sq_token(lexer: Lexer):
    string = StringIO("':1+1:1++:'ad:true:false:")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[0].text == "':1+1:1++:'ad"
    assert lexer.token_list[1].type == TokenType.Colon
    assert lexer.token_list[1].text == ":"
    assert lexer.token_list[2].type == TokenType.TRUE
    assert lexer.token_list[2].text == "true"
    assert lexer.token_list[3].type == TokenType.Colon
    assert lexer.token_list[3].text == ":"
    assert lexer.token_list[4].type == TokenType.FALSE
    assert lexer.token_list[4].text == "false"
    string = StringIO("'true:false'")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[0].text == "'true:false'"
    string = StringIO("'false:true'+1")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[0].text == "'false:true'"
    assert lexer.token_list[1].type == TokenType.Plus
    assert lexer.token_list[1].text == "+"
    assert lexer.token_list[2].type == TokenType.RealNumber
    assert lexer.token_list[2].text == "1"
    string = StringIO("' '")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[0].text == "' '"


def dq_token(lexer: Lexer):
    string = StringIO("\":1+1:1++:\"ad:true:false:")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[0].text == "\":1+1:1++:\"ad"
    assert lexer.token_list[1].type == TokenType.Colon
    assert lexer.token_list[1].text == ":"
    assert lexer.token_list[2].type == TokenType.TRUE
    assert lexer.token_list[2].text == "true"
    assert lexer.token_list[3].type == TokenType.Colon
    assert lexer.token_list[3].text == ":"
    assert lexer.token_list[4].type == TokenType.FALSE
    assert lexer.token_list[4].text == "false"
    string = StringIO("\"true:false\"")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[0].text == "\"true:false\""
    string = StringIO("\"false:true\"+1")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[0].text == "\"false:true\""
    assert lexer.token_list[1].type == TokenType.Plus
    assert lexer.token_list[1].text == "+"
    assert lexer.token_list[2].type == TokenType.RealNumber
    assert lexer.token_list[2].text == "1"
    string = StringIO("\" \"")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[0].text == "\" \""


def question_token(lexer: Lexer):
    string = StringIO("?")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Question
    assert lexer.token_list[0].text == "?"
    string = StringIO("a:?b")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Identifier
    assert lexer.token_list[0].text == "a"
    assert lexer.token_list[1].type == TokenType.Colon
    assert lexer.token_list[1].text == ":"
    assert lexer.token_list[2].type == TokenType.Question
    assert lexer.token_list[2].text == "?"
    assert lexer.token_list[3].type == TokenType.Identifier
    assert lexer.token_list[3].text == "b"
    # string = StringIO("a = 1? c:d")
    # lexer.tokenize(string)
    # assert lexer.token_list[0].type == TokenType.Identifier
    # assert lexer.token_list[0].text == "a"
    # assert lexer.token_list[1].type == TokenType.Blank
    # assert lexer.token_list[1].text == " "
    # assert lexer.token_list[2].type == TokenType.Assignment
    # assert lexer.token_list[2].text == "="
    # assert lexer.token_list[3].type == TokenType.Blank
    # assert lexer.token_list[3].text == " "


def comma_token(lexer: Lexer):
    string = StringIO(",")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Comma
    assert lexer.token_list[0].text == ","
    string = StringIO(", +,")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Comma
    assert lexer.token_list[0].text == ","
    assert lexer.token_list[1].type == TokenType.Blank
    assert lexer.token_list[1].text == " "
    assert lexer.token_list[2].type == TokenType.Plus
    assert lexer.token_list[2].text == "+"
    assert lexer.token_list[3].type == TokenType.Comma
    assert lexer.token_list[3].text == ","


def lp_rp_token(lexer: Lexer):
    string = StringIO("()")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Lp
    assert lexer.token_list[0].text == "("
    assert lexer.token_list[1].type == TokenType.Rp
    assert lexer.token_list[1].text == ")"

    string = StringIO("(1+2)")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Lp
    assert lexer.token_list[0].text == "("
    assert lexer.token_list[1].type == TokenType.RealNumber
    assert lexer.token_list[1].text == "1"
    assert lexer.token_list[2].type == TokenType.Plus
    assert lexer.token_list[2].text == "+"
    assert lexer.token_list[3].type == TokenType.RealNumber
    assert lexer.token_list[3].text == "2"
    assert lexer.token_list[4].type == TokenType.Rp
    assert lexer.token_list[4].text == ")"

    string = StringIO("(\"ass\" + \"1+2\")")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Lp
    assert lexer.token_list[0].text == "("
    assert lexer.token_list[1].type == TokenType.Identifier
    assert lexer.token_list[1].text == "\"ass\""
    assert lexer.token_list[2].type == TokenType.Blank
    assert lexer.token_list[2].text == " "
    assert lexer.token_list[3].type == TokenType.Plus
    assert lexer.token_list[3].text == "+"
    assert lexer.token_list[4].type == TokenType.Blank
    assert lexer.token_list[4].text == " "
    assert lexer.token_list[5].type == TokenType.Identifier
    assert lexer.token_list[5].text == "\"1+2\""
    assert lexer.token_list[6].type == TokenType.Rp
    assert lexer.token_list[6].text == ")"
    string = StringIO("('ass' + '1+2')")
    lexer.tokenize(string)
    assert lexer.token_list[0].type == TokenType.Lp
    assert lexer.token_list[0].text == "("
    assert lexer.token_list[1].type == TokenType.Identifier
    assert lexer.token_list[1].text == "'ass'"
    assert lexer.token_list[2].type == TokenType.Blank
    assert lexer.token_list[2].text == " "
    assert lexer.token_list[3].type == TokenType.Plus
    assert lexer.token_list[3].text == "+"
    assert lexer.token_list[4].type == TokenType.Blank
    assert lexer.token_list[4].text == " "
    assert lexer.token_list[5].type == TokenType.Identifier
    assert lexer.token_list[5].text == "'1+2'"
    assert lexer.token_list[6].type == TokenType.Rp
    assert lexer.token_list[6].text == ")"


def test_ge_and_gt():
    s1 = StringIO("dateDiff(getMinByLog($ *.InformedConsent.DSSTDAT), $ *.*.*, \"D\") > tt: true")
    l = Lexer()
    l.tokenize(s1)
    assert l.token_list[14].type == TokenType.Gt
    assert l.token_list[14].text == ">"
    assert "".join([token.text for token in l.token_list]) == s1.getvalue()
    s2 = StringIO("dateDiff(getMinByLog($ *.InformedConsent.DSSTDAT), $ *.*.*, \"D\")>=0: true")
    lexer = Lexer()
    lexer.tokenize(s2)
    assert lexer.token_list[13].type == TokenType.Ge
    assert lexer.token_list[13].text == ">="
    assert "".join([token.text for token in lexer.token_list]) == s2.getvalue()


def le_and_lt():
    s1 = StringIO("dateDiff(getMinByLog($ *.InformedConsent.DSSTDAT), $ *.*.*, \"D\") < tt: true")
    lexer = Lexer()
    lexer.tokenize(s1)
    assert lexer.token_list[14].type == TokenType.Lt
    assert lexer.token_list[14].text == "<"
    assert "".join([token.text for token in lexer.token_list]) == s1.getvalue()
    s2 = StringIO("dateDiff(getMinByLog($ *.InformedConsent.DSSTDAT), $ *.*.*, \"D\")<=0: true")
    lexer = Lexer()
    lexer.tokenize(s2)
    assert "".join([token.text for token in lexer.token_list]) == s2.getvalue()
    assert lexer.token_list[13].type == TokenType.Le
    assert lexer.token_list[13].text == "<="


def neq():
    s1 = StringIO("toDate($*.*.*)!=""&&toDate(getMinByLog($*.Informed Consent.DSSTDAT))")
    lexer = Lexer()
    lexer.tokenize(s1)
    assert lexer.token_list[4].type == TokenType.NEq
    assert lexer.token_list[4].text == "!="
    assert "".join([token.text for token in lexer.token_list]) == s1.getvalue()
    s1 = StringIO("toDate($*.*.*)!==""&&toDate(getMinByLog($*.Informed Consent.DSSTDAT))")
    lexer = Lexer()
    lexer.tokenize(s1)
    assert lexer.token_list[4].type == TokenType.Nidentity
    assert lexer.token_list[4].text == "!=="
    assert "".join([token.text for token in lexer.token_list]) == s1.getvalue()


def assign():
    s1 = StringIO("a=1")
    lexer = Lexer()
    lexer.tokenize(s1)
    assert lexer.token_list[1].type == TokenType.Assignment
    assert lexer.token_list[1].text == "="
    assert "".join([token.text for token in lexer.token_list]) == s1.getvalue()
    s1 = StringIO("(a=1)")
    lexer = Lexer()
    lexer.tokenize(s1)
    assert lexer.token_list[2].type == TokenType.Assignment
    assert lexer.token_list[2].text == "="
    assert "".join([token.text for token in lexer.token_list]) == s1.getvalue()
    s1 = StringIO("(a=1)&&b=1")
    lexer = Lexer()
    lexer.tokenize(s1)
    assert lexer.token_list[2].type == TokenType.Assignment
    assert lexer.token_list[7].type == TokenType.Assignment
    assert lexer.token_list[2].text == "="
    assert lexer.token_list[7].text == "="
    assert "".join([token.text for token in lexer.token_list]) == s1.getvalue()


def eq():
    s1 = StringIO("a==1")
    lexer = Lexer()
    lexer.tokenize(s1)
    assert lexer.token_list[1].type == TokenType.Eq
    assert lexer.token_list[1].text == "=="
    assert "".join([token.text for token in lexer.token_list]) == s1.getvalue()
    s1 = StringIO("(a==1)")
    lexer = Lexer()
    lexer.tokenize(s1)
    assert lexer.token_list[2].type == TokenType.Eq
    assert lexer.token_list[2].text == "=="
    assert "".join([token.text for token in lexer.token_list]) == s1.getvalue()
    s1 = StringIO("(a==1)&&b==1")
    lexer = Lexer()
    lexer.tokenize(s1)
    assert lexer.token_list[2].type == TokenType.Eq
    assert lexer.token_list[7].type == TokenType.Eq
    assert lexer.token_list[2].text == "=="
    assert lexer.token_list[7].text == "=="
    assert "".join([token.text for token in lexer.token_list]) == s1.getvalue()


def eq_plus():
    s1 = StringIO("a=+1")
    lexer = Lexer()
    lexer.tokenize(s1)
    assert lexer.token_list[1].type == TokenType.EqPlus
    assert lexer.token_list[1].text == "=+"
    assert "".join([token.text for token in lexer.token_list]) == s1.getvalue()
    s1 = StringIO("(a=+1)")
    lexer = Lexer()
    lexer.tokenize(s1)
    assert lexer.token_list[2].type == TokenType.EqPlus
    assert lexer.token_list[2].text == "=+"
    assert "".join([token.text for token in lexer.token_list]) == s1.getvalue()
    s1 = StringIO("(a=+1)&&b=+1")
    lexer = Lexer()
    lexer.tokenize(s1)
    assert lexer.token_list[2].type == TokenType.EqPlus
    assert lexer.token_list[7].type == TokenType.EqPlus
    assert lexer.token_list[2].text == "=+"
    assert lexer.token_list[7].text == "=+"
    assert "".join([token.text for token in lexer.token_list]) == s1.getvalue()


def eq_minus():
    s1 = StringIO("a=-1")
    lexer = Lexer()
    lexer.tokenize(s1)
    assert lexer.token_list[1].type == TokenType.EqMinus
    assert lexer.token_list[1].text == "=-"
    assert "".join([token.text for token in lexer.token_list]) == s1.getvalue()
    s1 = StringIO("(a=-1)")
    lexer = Lexer()
    lexer.tokenize(s1)
    assert lexer.token_list[2].type == TokenType.EqMinus
    assert lexer.token_list[2].text == "=-"
    assert "".join([token.text for token in lexer.token_list]) == s1.getvalue()
    s1 = StringIO("(a=-1)&&b=-1")
    lexer = Lexer()
    lexer.tokenize(s1)
    assert lexer.token_list[2].type == TokenType.EqMinus
    assert lexer.token_list[7].type == TokenType.EqMinus
    assert lexer.token_list[2].text == "=-"
    assert lexer.token_list[7].text == "=-"
    assert "".join([token.text for token in lexer.token_list]) == s1.getvalue()


def aways_eq_plus():
    s1 = StringIO("a===1")
    lexer = Lexer()
    lexer.tokenize(s1)
    assert lexer.token_list[1].type == TokenType.AwaysEq
    assert lexer.token_list[1].text == "==="
    assert "".join([token.text for token in lexer.token_list]) == s1.getvalue()
    s1 = StringIO("(a===1)")
    lexer = Lexer()
    lexer.tokenize(s1)
    assert lexer.token_list[2].type == TokenType.AwaysEq
    assert lexer.token_list[2].text == "==="
    assert "".join([token.text for token in lexer.token_list]) == s1.getvalue()
    s1 = StringIO("(a===1)&&b===1")
    lexer = Lexer()
    lexer.tokenize(s1)
    assert lexer.token_list[2].type == TokenType.AwaysEq
    assert lexer.token_list[7].type == TokenType.AwaysEq
    assert lexer.token_list[2].text == "==="
    assert lexer.token_list[7].text == "==="
    assert "".join([token.text for token in lexer.token_list]) == s1.getvalue()


def mul_and_divide():
    string = StringIO("toNum($*.*.EGORRES.881)/1000")
    lexer = Lexer()
    lexer.tokenize(string)

    assert "".join([token.text for token in lexer.token_list]) == string.getvalue()
    string = StringIO("toNum($*.*.EGORRES.881)*1000")
    lexer = Lexer()
    lexer.tokenize(string)

    assert "".join([token.text for token in lexer.token_list]) == string.getvalue()



def test_vars():
    s = StringIO("$*.*.EGPERF==\"Y\"")
    lexer = Lexer()
    lexer.tokenize(s)
    assert lexer.token_list[0].type == TokenType.Variable
    assert lexer.token_list[0].text == "$*.*.EGPERF"
    assert lexer.token_list[1].type == TokenType.Eq
    assert lexer.token_list[1].text == "=="
    assert lexer.token_list[2].type == TokenType.Identifier
    assert lexer.token_list[2].text == "\"Y\""
    s = StringIO("$C1D1.ONC-392 Administration.ECSTTIM+\":00\"")
    lexer = Lexer()
    lexer.tokenize(s)
    assert lexer.token_list[0].type == TokenType.Variable
    assert lexer.token_list[1].type == TokenType.Plus
    assert lexer.token_list[2].type == TokenType.Identifier
    s = StringIO("$*.*.*>=\":00\"")
    lexer = Lexer()
    lexer.tokenize(s)
    assert lexer.token_list[0].type == TokenType.Variable
    assert lexer.token_list[1].type == TokenType.Ge
    assert lexer.token_list[2].type == TokenType.Identifier
    s = StringIO("$C1D1.ONC-392 Administration.ECSTTIM.1992+\":00\"")
    lexer = Lexer()
    lexer.tokenize(s)
    assert lexer.token_list[0].type == TokenType.Variable
    assert lexer.token_list[1].type == TokenType.Plus
    assert lexer.token_list[2].type == TokenType.Identifier
    s = StringIO("$*.*.*.1992")
    lexer = Lexer()
    lexer.tokenize(s)
    assert lexer.token_list[0].type == TokenType.Variable
    s = StringIO("$*.*.*.1992>=\":00\"")
    lexer = Lexer()
    lexer.tokenize(s)
    assert lexer.token_list[0].type == TokenType.Variable
    assert lexer.token_list[1].type == TokenType.Ge
    assert lexer.token_list[2].type == TokenType.Identifier


def test_math_name():
    s = StringIO("Math.power()")
    l = Lexer()
    l.tokenize(s)
    print(l.token_list)






class Squeue:
    def __init__(self):
        self._data = []

    def append(self, e):
        self._data.append(e)

    @property
    def data(self) -> list:
        return self._data

    def get(self, i):
        return self._data[i]

    def top(self):
        if self.empty(): return None
        return self._data[0]

    def count(self): return len(self._data)

    def empty(self): return self.count() == 0
